
// Geometric deduction rules, as described in Section 5 of the
// Journal of Logic and Computation paper. 
// Written by Hans de Nivelle, Jan-Feb 2015.

#include "deduction.h"
#include "unification.h"


void geometric::hornchain::push_back( const inference& inf )
{
   if( auto* p = dynamic_cast< const userule* > ( inf. getcontents( )))
   {
      const inference& sub = p -> r. getinference( );
      if( auto* q = dynamic_cast< const hornchain* > ( sub. getcontents( )))
      {
         // Sub is also a Horn chain.

         for( const inference& i : q -> parents )
            parents. push_back( i );
         return; 
      }
   }
     
   // Fall through means that we insert inf itself:

   parents. push_back( inf );
}

#if 0
void geometric::hornchain::print( std::ostream& out ) const
{
   out << "hornres( ";
   for( auto p = parents. begin( ); p != parents. end( ); ++ p )
   {
      if( p != parents. begin( ))
         out << ", ";
      p -> print( out );
   }
   out << " )";
}
#endif  

void geometric::hornchain::print( std::ostream& out ) const
{
   for( auto p = parents. begin( ); p != parents. end( ); ++ p )
   {
      if( p != parents. begin( ))
         out << '+';
      p -> print( out );
   }
}

 
void geometric::disjres::print( std::ostream& out ) const
{
   out << "disj( ";
   for( auto p = parents. begin( ); p != parents. end( ); ++ p )
   {
      if( p != parents. begin( ))
         out << ", ";
      p -> print( out ); 
   }
   out << " )";
}


void geometric::existsres::print( std::ostream& out ) const
{
   out << "exists( ";
   existential. print( out );
   out << ", ";
   lemma. print( out );
   out << " )";
}


//////////////////////////////////////////////////////////////////

geometric::varatom geometric::replace( varatom at, variable v1, variable v2 )
{
   for( auto& arg : at )
   {
      if( arg == v1 )
         arg = v2;
   }
   return at; 
}


bool geometric::groundconflict( const groundatom& at1, const groundatom& at2 )
{
   return at1.p == at2.p && 
          ( at1.t & at2.t ) == truthval::empty &&
          equal_arguments( at1, at2 );
}

bool geometric::iscreate( element e, const groundatom& at )
{
   return at.p.iscreate( ) && 
          at.t == truthval::F &&
          at[0] == e;
}

bool geometric::isequalityfrom( element e, const groundatom& at ) 
{
   return at.p.isequality( ) && 
          at.t == truthval::T &&
          at[0] == e && at[1] != e;
}

bool geometric::isequalityinto( element e, const groundatom& at )
{
   return at.p.isequality( ) && 
          at.t == truthval::T &&
          at[1] == e && at[0] != e;
}


size_t 
geometric::find( size_t val, const std::vector<size_t> & vect ) 
{
   size_t j = 0; 
   while( j < vect. size( )) 
   {
      if( val == vect[j] )
         return j;
      ++ j;
   }
   return j;
}


namespace
{
   using namespace geometric;

   void transfer( const subst< element > & orig,
                  label lab,
                  const unifier& unif,
                  normalizer& norm,
                  const set< variable > & varsinresult,
                  subst< element > & result )
   {
      for( std::pair<variable,element> a : orig )
      {
         a. first = unif. apply( norm, a. first, lab );
         // std::cout << a. first << " -> " << a. second << "\n";
         if( varsinresult. contains( a.first ))
            result. assign( a.first, a.second );
      }
   }
}

geometric::matched< geometric::disjunctive >
geometric::ex_resolve( const matched<existential> & exists,
                       element skolem_constant,
                       const false_instance& inst_confl ) 
{

   if( SHOWDEDUCTION )
   {
      std::cout << "\n";
      std::cout << "Existential resolution:\n";
      std::cout << exists << "\n";
      std::cout << "skolem constant " << skolem_constant << "\n";
      std::cout << "with " << inst_confl << "\n\n";
   }

   variable exvar = exists.r. getvariable( );
   const varatom& exatom = exists.r. getatom( );

   // All free variables of the existential formula must have an 
   // assignment, and this assignment must be not the skolem_constant.
   // The existential variable must not have an assignment.

   {
      set< variable > freevars = exists.r. getlhsvariables( );
 
      for( variable v : freevars ) 
      {
         auto p = exists. inst. clookup(v);
 
         if( !p )
            throw std::runtime_error( "existential resolution: unassigned "\
                                      "variable in existential formula" );
         if( *p == skolem_constant )
            throw std::runtime_error( "existential resolution: variables "\
                                      "different from the existential "\
                                      "variable cannot be mapped into "\
                                      "the Skolem constant" );
      } 
   }

   {
      auto p = exists. inst. clookup( exvar );
      if(p)
         throw std::runtime_error( "existential resolution: existential "\
                                   "variable has assignment" );

      if( !exatom. contains( exvar ))
         throw std::runtime_error( "existential resolution: the existential "\
                                   "variable does not occur in the "\
                                   "existential atom" );
   }

   // The existential atom cannot be an equality atom. 
   // Create seems to be possible, so we allow it. 

   if( exatom.p. isequality( ))
   {
      throw std::runtime_error( "existential resolution: existential atom "\
                                "is equality" );
   }

   groundatom exatomground = 
                   exists. inst. apply( exatom, exvar, skolem_constant );

   // We can start looking at the conflict formula: 

   if( !inst_confl.r.isdisjunctive( ))
      throw std::runtime_error( "existential resolution: "\
                                "conflict formula must be disjunctive" );

   const auto& confl = inst_confl.r.getdisjunctive( ); 

   // Every variable in the conflict formula must have an assignment: 

   set<variable> freevars_confl = confl. getlhsvariables( );

   for( variable v : freevars_confl )
   {
      auto p = inst_confl.inst.clookup(v);
      if( !p )
         throw std::runtime_error( "existential resolution: unassigned "\
                                      "variable in conflict formula" );
   }

   // The conflict formula must not have a conclusion:

   if( confl. rhs_size( ) != 0 )
   {
      throw std::runtime_error( "existential resolution: conflict formula "\
                                "has non-empty right hand side" ); 
   }

   unifier unif;
   label lab_exists = label(1);
   label lab_confl  = label(2);

   // We start by constructing the most general unifier:
   // Every variable in the conflict formula that is mapped into the 
   // Skolem constant, must be unified with the existential variable.

   for( variable v : freevars_confl )
   {
      if( *inst_confl.inst.clookup(v) == skolem_constant )
      {
         unif. unify( exvar, lab_exists, v, lab_confl );
      }
   }

   // The atoms in confl, whose ground instances are in conflict 
   // with the ground instance of the existential atom, 
   // must also be unified, because they will resolve. 

   for( const varatom& at : confl. get_lhs( ))
   {
      groundatom inst = inst_confl.inst.apply(at);
      if( groundconflict( inst, exatomground ))
      {
         unif. unify_arguments( exatom, lab_exists, at, lab_confl );
      }
   }

   normalizer norm;

   // We can now start constructing the result: 

   disjunctive result;

   // Substitute and copy atoms from the existential formula:

   for( const varatom& at : exists.r. get_lhs( ))
   {
      result. insert_lhs( unif. apply( norm, at, lab_exists ));
   }

   // Apply the unifier on the existential atom. We need it when we
   // encounter equalities. 

   varatom exatomunified = unif. apply( norm, exatom, lab_exists );   

   for( const varatom& at : confl. get_lhs( ))
   {
      groundatom inst = inst_confl.inst.apply( at ); 
      varatom unified = unif. apply( norm, at, lab_confl );
         // We now have a ground instance, and a unified instance of at.

      if( !inst. contains( skolem_constant ))
      {
         // Atoms whose instantiation does not contain the Skolem constant,
         // are copied into the left hand side: 

         result. insert_lhs( unified ); 
         goto done;
      }

      if( isequalityfrom( skolem_constant, inst ))
      {
         result. insert_rhs( replace( exatomunified, unified[0], unified[1] ));
         goto done;
      }

      if( isequalityinto( skolem_constant, inst ))
      {
         result. insert_rhs( replace( exatomunified, unified[1], unified[0] )); 
         goto done;
      }

      if( iscreate( skolem_constant, inst ))
         goto done;

      if( groundconflict( inst, exatomground ))
         goto done;  

      throw std::runtime_error( "existential resolution: atom "\
                                "contains Skolem element, but is not "\
                                "resolvable" );
   done:
      ;
   }

   set<variable> freevars_result = result. getlhsvariables( );


   subst< element > inst; 
   transfer( exists. inst, lab_exists, unif, norm, 
                freevars_result, inst );
   transfer( inst_confl. inst, lab_confl, unif, norm, 
                freevars_result, inst ); 
   
   // We put 0 choices in the false instance, because we have no way  
   // of determining the real number without knowing the interpretation.

   if( SHOWDEDUCTION )
   {
      std::cout << "The result is:\n";
      std::cout << result << "\n\n";
   }

   return { result, inst }; 

}


geometric::matched< geometric::disjunctive > 
geometric::disj_resolve( 
      const matched< disjunctive > & disj, 
      const std::vector<size_t> & extensions, 
      const std::vector< false_instance > & inst_conflicts ) 
{

   if( SHOWDEDUCTION )
   {
      std::cout << "\n";
      std::cout << "Disjunction resolution:\n";
      std::cout << "main disjunction:\n";
      std::cout << disj << "\n";

      std::cout << "extensions: {";
      for( size_t i = 0; i != extensions. size( ); ++ i )
      {
         if( i != 0 )
            std::cout << ", ";
         else
            std::cout << " ";
         std::cout << extensions[i];
      }
      std::cout << " }\n\n";
 
      std::cout << "conflict formulas:\n";
      for( const auto& confl : inst_conflicts )
      {
         std::cout << confl << "\n";
      }
      std::cout << "\n";
   }

   // Every variable in the disjunctive formula must have an assignment:

   {
      auto s = disj. getunassigned( );
      if( s. size( )) 
      {
         std::cerr << s << "\n";
         throw std::runtime_error( "disjunctive resolution: there are "\
                           "unassigned variables in the main disjunction" );
      } 
   }

   for( const auto& p : inst_conflicts )
   {
      if( !p.r. isdisjunctive( ))
         throw std::runtime_error( "disjunctive resolution: "\
                                   "one of the conflict formulas "\
                                   "is not disjunctive" );

      const disjunctive& confl = p.r.getdisjunctive( );

      set< variable > freevars = confl. getlhsvariables( );
      for( variable v : freevars )
      {
         if( !p.inst.clookup(v) )
            throw std::runtime_error( "disjunctive resolution: unassigned "\
                          "variable in one of the conflict formulas" );
      }

      if( confl. rhs_size( ))
         throw std::runtime_error( "disjunctive resolution: One of the "\
                              "conflict formulas has nonempty rhs" );  
   }

   for( size_t i : extensions )
   {
      if( i >= disj.r. rhs_size( ))
         throw std::runtime_error( "disjunctive:resolution: index "\
                      "of active atom is too large" );
   }

   if( extensions. size( ) < inst_conflicts. size( ))
      throw std::runtime_error( "disjunctive resolution: There are "\
                                "too many conflict formulas" );

   unifier unif;
   label lab_disj = label(-1);

   // In the first pass of the active atoms, we construct the unifier:

   for( size_t e = 0; e < inst_conflicts. size( ); ++ e )
   {
      size_t j = extensions[e];
      groundatom inst = disj. inst. apply( disj.r[j] );
      label current_label = label(j);

      const disjunctive& confl = inst_conflicts[e]. r. getdisjunctive( );
      for( const varatom& at : confl. get_lhs( ))
      {
         groundatom atinst = inst_conflicts[e]. inst. apply( at );
         if( groundconflict( inst, atinst ))
         {
            unif. unify_arguments( disj.r[j], lab_disj, at, current_label );
         }
      }
   }

   normalizer norm;

   disjunctive result;
      // Substitute and copy the passive atoms from the disjunctive formula:

   for( const varatom& at : disj.r. get_lhs( ))
   {
      result. insert_lhs( unif. apply( norm, at, lab_disj ));
   }
  
   for( size_t j = 0; j < disj. r. rhs_size( ); ++ j )
   {
      size_t position = find( j, extensions );
      if( position < inst_conflicts. size( ))
      {
         // If j is in S+, and a conflict formula is present,
         // then disj.r.[j] gets resolved away: 

         groundatom inst = disj. inst. apply( disj.r[j] );
         label current_label = label(j);

         const disjunctive& confl = 
                  inst_conflicts[ position ]. r. getdisjunctive( );

         for( const varatom& at : confl. get_lhs( ))
         {
            groundatom atinst = inst_conflicts[position].inst.apply( at );
            if( !groundconflict( inst, atinst ))
            {
               result. insert_lhs( unif. apply( norm, at, current_label )); 
            }
         }
      }
      else
      {
         if( position < extensions. size( ))
         {
            // Extension that won't resolve. It goes to the right hand side:

            result. insert_rhs( unif. apply( norm, disj.r[j], lab_disj )); 
         }
         else
         {
            // Immediate conflict (S- in the paper). It goes to the
            // left hand side: 
 
            result. insert_lhs( unif. apply( norm, disj.r[j], lab_disj ));
         }
      }
   }

   set<variable> freevars_result = result. getlhsvariables( );

   subst< element > resinst; 
   transfer( disj. inst, lab_disj, unif, norm, freevars_result, resinst );

   for( size_t j = 0; j < disj.r. rhs_size( ); ++ j )
   {
      label current_label = label(j);

      auto position = find( j, extensions );
      if( position < inst_conflicts. size( ))
      {
         transfer( inst_conflicts[position]. inst, current_label, 
                   unif, norm,
                   freevars_result, resinst ); 
      }
   }

   if( SHOWDEDUCTION )
   {
      std::cout << "The result is:\n";
      std::cout << result << "\n";
      std::cout << resinst << "\n\n";
   }

   return { result, resinst };

}




std::ostream& geometric::operator << ( std::ostream& out, merging m )
{
   out << m. v1 << " == " << m. v2;
   return out;
}


void geometric::mergings::print( std::ostream& out ) const
{
   out << "mergings( ";
   for( auto p = m. begin( ); p != m. end( ); ++ p )
   {
      if( p != m. begin( ))
         out << ", ";
      out << *p; 
   }
   out << "; ";
   parent. print( out ); 
   out << " )";
}


void geometric::mergings::flatten( ) 
{
   if( auto* p = dynamic_cast< const userule* > ( parent. getcontents( )))
   {
      const inference& sub = p -> r. getinference( );

      if( auto* q = dynamic_cast< const mergings* > ( sub. getcontents( )))
      {
         // We construct a new set of mergings by appending
         // our mergings to the mergings in *q:

         std::vector< merging > newmerge = q -> m;
         for( auto ours : m )
            newmerge. push_back( ours ); 

         parent = q -> parent;
         m = std::move( newmerge ); 
      }
   }

   // Fall through means that we do nothing. 
}


namespace 
{
   // Apply the chain of mergings (always v1 ==> v2) on at:

   void apply( const std::vector< merging > & eq, varatom& at )
   {
      for( merging m : eq )
      {
         for( variable& v : at )
         {
            if( v == m.v1 )
               v = m.v2;
         }
      }
   }
}


geometric::matched<geometric::disjunctive>
geometric::merge( const matched<disjunctive> & disj,
                  const std::vector<merging> & eq )
{
   if( SHOWDEDUCTION )
   {
      std::cout << "\n\n";
      std::cout << "Merging in " << disj << "\n";
      std::cout << "mergings: {";
      for( auto p = eq. begin( ); p != eq. end( ); ++ p )
      {
         if( p != eq. begin( )) 
            std::cout << ", ";
         else
            std::cout << " ";
         std::cout << *p;
      }  
      std::cout << " }\n";
      std::cout << "\n";
   }

   matched< disjunctive > result( {}, { disj. inst } );

   for( varatom at : disj. r. get_lhs( ))
   {
      apply( eq, at ); 
      result. r. insert_lhs( at );
   }

   for( varatom at : disj. r. get_rhs( ))
   {
      apply( eq, at );
      result. r. insert_rhs( at );
   }

   if( SHOWDEDUCTION )
   {
      std::cout << "The result is:\n";
      std::cout << result << "\n\n";
   }

   return result;  
}


auto geometric::normalize( const existential& exists ) -> existential
{
   normalizer norm;
   label lab(0);

   if( SHOWDEDUCTION )
   {
      std::cout << "\n";
      std::cout << "Normalizing:\n";
      std::cout << exists << "\n\n";
   }

   // We have to go one time through the left hand side because we have
   // to call the constructor with the existential formula and the
   // quantified variable.

   for( const varatom& at : exists. get_lhs( ))
      norm. apply( at, lab );

   if( norm. contains( exists. getvariable( ), lab ))
   {
      throw std::runtime_error( "normalize existential: "\
                    "existential variable occurs in the left hand side" );    
 
   }

   variable v = norm. lookup( exists. getvariable( ), lab );
   varatom at = norm. apply( exists. getatom( ), lab );

   existential result( v, at );

   for( const varatom& at: exists. get_lhs( ))
      result. insert_lhs( norm. apply( at, lab ));

   if( SHOWDEDUCTION )
   {
      std::cout << "Result is " << result << "\n\n";
   }

   return result;
}


auto geometric::normalize( const disjunctive& disj ) -> disjunctive
{
   normalizer norm;
   label lab(0);

   if( SHOWDEDUCTION )
   {
      std::cout << "\n";
      std::cout << "Normalizing:\n";
      std::cout << disj << "\n\n";
   }

   disjunctive result;

   for( const varatom& at: disj. get_lhs( ))
      result. insert_lhs( norm. apply( at, lab ));

   for( const varatom& at: disj. get_rhs( ))
      result. insert_rhs( norm. apply( at, lab ));

   if( SHOWDEDUCTION )
   {
      std::cout << "Result is " << result << "\n\n";
   }

   return result;
}


