
// Written by Hans de Nivelle, Jan-Feb 2015.
// Part of Geo III.

#ifndef GEOMETRIC_DEDUCTION_INCLUDED
#define GEOMETRIC_DEDUCTION_INCLUDED  1

#include "rule.h"
#include "false_instance.h"
#include "matched.h"


namespace geometric
{
   constexpr bool SHOWDEDUCTION = false;

   // Inference can start from existing rules:

   struct userule : infbase
   {
      rule r;

      userule( const rule& r ) 
         : r{r}
      { }

      userule* clone( ) const& override 
         { return new userule( *this ); }

      userule* clone( ) && override 
         { return new userule( std::move( *this )); }

      size_t nrparents( ) const override { return 0; }
      virtual const inference& operator [] ( size_t ) const override
         { throw std::runtime_error( "userule has no parents" ); }

      void print( std::ostream& out ) const override
         { r. printnumber( out ); } 

   };


   // A Horn chain is a chain of Horn resolution steps. 
   // Each parent should be able to resolve with its neighbours.

   struct hornchain : infbase
   {
      std::vector< inference > parents;

      hornchain( const inference& left, const inference& right ) 
      {
         push_back( left );
         push_back( right );
            // If left/right is also obtained by hornchain,
            // then the parents of left/right are added instead
            // of left/right itself. 
      }

      hornchain* clone( ) const& override 
         { return new hornchain( *this ); }

      hornchain* clone( ) && override
         { return new hornchain( std::move( *this )); }

      size_t nrparents( ) const { return parents. size( ); }
      const inference& operator [] ( size_t i ) const override
         { return parents[i]; }

      void print( std::ostream& ) const override;

   private:
      void push_back( const inference& inf ); 
   };


   // A disjunction resolution step resolves one disjunction 
   // (the first parent) with a lot of side clauses, where each side 
   // clause resolves away one of the active atoms of the first parent. 

   struct disjres : infbase 
   {
      std::vector< inference > parents;

      disjres( std::initializer_list< inference > par )
         : parents{ par }
      { }

      disjres( const std::vector< inference > & parents )
         : parents{ parents }
      { }

      disjres( std::vector< inference > && parents )
         : parents{ std::move( parents ) }
      { }

      disjres* clone( ) const& override
         { return new disjres( *this ); }

      disjres* clone( ) && override
         { return new disjres( std::move( *this )); }

      size_t nrparents( ) const { return parents. size( ); }
      const inference& operator [] ( size_t i ) const override
         { return parents[i]; }

      void print( std::ostream& ) const override;
   };


   // Existential resolution is binary, between an existential formula,
   // and a lemma. 
 
   struct existsres : infbase
   {
      inference existential; 
      inference lemma; 

      existsres( const inference& existential, const inference& lemma )
         : existential{ existential },
           lemma{ lemma }
      { }

      existsres* clone( ) const& override 
         { return new existsres( *this ); } 
      existsres* clone( ) && override
         { return new existsres( std::move( *this )); }

      size_t nrparents( ) const { return 2; }
      const inference& operator[] ( size_t i ) const override
      {
         if( i == 0 )
            return existential;
         else
            return lemma;  
      }

      void print( std::ostream& out ) const override;
   };


   // Return at[ v1 ==> v2 ].

   varatom replace( varatom at, variable v1, variable v2 );

   bool groundconflict( const groundatom& at1, const groundatom& at2 );
      // True if at1 and at2 are in ground conflict.

   bool iscreate( element e, const groundatom& at );
      // True if at has from #_{F} e.

   bool isequalityfrom( element e, const groundatom& at );
      // True if at has form e = e', where e' is not e. 
   bool isequalityinto( element e, const groundatom& at );
      // True if at has form e' = e, where e' is not e.

   size_t find( size_t val, const std::vector<size_t> & vect );
      // Return index on which val occurs in vector, if it does.
      // Otherwise vect. size( ).


   // skolem_constant is the instance of the existential variable
   // in inst_exists.

   matched<disjunctive>
   ex_resolve( const matched<existential> & exists, 
               element skolem_constant, 
               const false_instance& inst_lemma );

   // extensions is a vector of indices of rhs atoms of inst_disj.
   // (These are the rhs atoms by which the interpretation can
   //  be extended. (Set S+). The others are in set (S-)). 
   //  The order of extensions is not fixed.
   // For an initial segment of extensions, there may be 
   // conflict formulas. These will be resolved with the corresponding  
   // extensions. The remaining extensions will stay in the right hand side. 
   // Right hand side atoms whose index does not occur in extensions,
   // are moved to the left hand side.

   matched<disjunctive> 
   disj_resolve( const matched<disjunctive> & disj, 
      const std::vector<size_t> & extensions, 
      const std::vector< false_instance > & inst_conflicts ); 

   struct merging
   {
      variable v1;
      variable v2;

      merging( variable v1, variable v2 )
         : v1{v1}, v2{v2}
      { }
   };

   std::ostream& operator << ( std::ostream& , merging );



   struct mergings : infbase
   {
      inference parent;
      std::vector< merging > m;

      mergings( const inference& parent,
                const std::vector<merging> & m )
         : parent{ parent },
           m{m}
      { 
         flatten( ); 
      }

      mergings( const inference& parent,
                std::vector<merging> && m )
         : parent{ parent },
           m{ std::move(m) }
      { 
         flatten( );
      }

      mergings* clone( ) const& override 
         { return new mergings( *this ); }

      mergings* clone( ) && override
         { return new mergings( std::move( *this )); }

      size_t nrparents( ) const override { return 1; }

      const inference& operator [] ( size_t i ) const 
      {
         if( i == 0 )
            return parent;
         else
            throw std::runtime_error( "merging has only one parent" );  
      }

      void print( std::ostream& ) const override;

   private: 
      void flatten( );
   };


   matched<disjunctive> merge( const matched<disjunctive> & disj,
                               const std::vector<merging> & eq );
      // Merge does not normalize, so that we will keep the original
      // variables. This is useful for proof logging, when we 
      // make a couple of consecutive mergings. Otherwise, one would  
      // loose track of the variables due to normalizations. 

   existential normalize( const existential& exists );
   disjunctive normalize( const disjunctive& disj );
   
}

#endif

