
// Written by Hans de Nivelle, August 2014.
// Rewritten in May 2015.

#ifndef GEOMETRIC_DISJUNCTIVE_INCLUDED
#define GEOMETRIC_DISJUNCTIVE_INCLUDED  1

#include "formula.h"

namespace geometric
{

   // Disjunctive formulas can be generated in two ways:
   // 1. Initial formulas. It is essential for completeness that
   //    the disjunction is normal, which means that every non-active
   //    atom is in the rhs. First use one of the constructors 
   //    to construct the disjunction, putting all atoms in the lhs.
   //    After that, call set_rhs( ), makerangerestricted( ), sort_lhs( ).
   //
   // 2. Formulas derived by resolution. If the parents were range
   //    restricted, then the new formula is automatically range restricted
   //    as well. If any non-passive atoms occur in the lhs, they should 
   //    remain there, because they are in conflict with the interpretation.  
   //    Call removeredundant( ), sort_lhs( ). 

   class disjunctive : public formula
   {
      std::vector< varatom > rhs;
         // Contains only atoms of form p_{E} or p_{T}. 

   public: 
      // The constructors use formula::insert_lhs, which merges
      // as much as possible.
 
      disjunctive( ) = default; 
      disjunctive( std::initializer_list< varatom > lhs )
         : formula{ lhs } { } 
      disjunctive( const std::vector< varatom > & lhs )
         : formula{ lhs } { } 

      void insert_rhs( const varatom& at );
         // Insert the atom into the rhs if it does not occur already. 
         // The atom must be not passive. Atoms with truthvalue ET
         // are not split automatically. 
 
      void set_rhs( );
         // Decide which atoms should be in the rhs. These are 
         // all non-equality,non-create atoms, whose truthvalue
         // does not contain F. 
         // If the truthvalue equals ET, then the atom is split into
         // two atoms.

      void makerangerestricted( ) override; 
         // Make the disjunction range-restricted by inserting 
         // suitable #_{F} predicates into the lhs.
         // All polymorphic functions that are defined for formula,
         // act as if the rhs equals FALSE.

      using formula::isfree;          // Crazy, see BS, Section 20.3.5. 
      bool hasothervars( variable v ) const override;

      size_t rhs_size( ) const { return rhs. size( ); }
      const varatom& operator [] ( size_t j ) const
         { return rhs[j]; } 
      const std::vector< varatom > & get_rhs( ) const 
         { return rhs; } 

      bool isfree( variable v, 
             std::vector< varatom > :: const_iterator it ) const override;

      void print( std::ostream& out ) const override; 

      disjunctive* clone( ) const & override 
         { return new disjunctive( *this ); }
      disjunctive* clone( ) && override 
         { return new disjunctive( std::move( *this )); }

   };

}

#endif

