
#include "false_instance.h"

namespace geometric
{

   std::ostream& operator << ( std::ostream& out, const false_instance& f )
   {
      out << "false instance:\n";
      out << f. r << "\n";
      out << "number of choices: " << f. nrchoices << "\n";
      out << f. inst << "\n";
      out << "assumptions: " << f. assumptions << "\n";
      return out;
   }


namespace
{
 
   // at1(inst1) subsumes at2(inst2) must mean:
   // Whenever at2(inst2) conflicts an interpretation I, then so does
   // at1(inst1). 


   bool subsumes( const varatom& at1, const subst<element> & inst1,
                  const varatom& at2, const subst<element> & inst2 )
   {

      if( at1.p. iscreate( ) && at1.t == truthval::F ) 
      {
         element e = *inst1. lookup( at1[0] );
    
         for( auto v : at2 )
         {  
            if( e == *inst2. lookup(v))
               return true;
         }

         return false;  
      }

      if( at1.p. isequality( ))
      {
         if( at1.p != at2.p || at1.t != at2.t ) 
            return false;

         auto e1 = *inst1. lookup( at1[0] );
         auto e2 = *inst1. lookup( at1[1] );
       
         auto f1 = *inst2. lookup( at2[0] );
         auto f2 = *inst2. lookup( at2[1] );

         return ( f1 == f2 ) || 
                ( e1 == f1 && e2 == f2 ) || ( e1 == f2 && e2 == f1 );
      }

      // A usual, boring predicate symbol, without any special properties:

      if( at1.p != at2.p )
         return false;

      if( ( at1.t & at2.t ) != at1.t )
         return false;

      auto p1 = at1. args_cbegin( );
      auto p2 = at2. args_cbegin( );
      while( p1 != at1. args_cend( ))
      {   
         auto e1 = *inst1. lookup( *p1 );
         auto e2 = *inst2. lookup( *p2 );
        
         if( e1 != e2 ) return false;
         ++ p1; ++ p2;
      }
      
      return true;  
   }


   bool 
   subsumes( variable ex1, const varatom& at1, const subst<element> & inst1,
             variable ex2, const varatom& at2, const subst<element> & inst2 )
   {

      // We use the fact that existentially quantified atoms cannot be 
      // create or equality. In addition, they have only simple 
      // truth-values. 

      if( at1.p != at2.p || at1.t != at2.t ) 
         return false;
 
      auto p1 = at1. args_cbegin( );
      auto p2 = at2. args_cbegin( );
      while( p1 != at1. args_cend( ))
      {
         // If one of *p1,*p2 equals its existential variable,
         // then both must equal their existential variable.

         if( *p1 == ex1 || *p2 == ex2 )
         {
            if( *p1 != ex1 || *p2 != ex2 )
               return false;
         }
         else
         {
            // Otherwise, the looked up values must be equal:

            auto e1 = *inst1. lookup( *p1 );
            auto e2 = *inst2. lookup( *p2 );

            if( e1 != e2 )
               return false;
         }
         ++ p1; ++ p2;
      }

      return true;
   }


   bool implies( const varatom& at1, const subst<element> & inst1,
                 const false_instance& ff ) 
   {
#if 0
      std::cout << "implies:\n";
      std::cout << at1 << "\n";
      std::cout << inst1 << "\n";
      std::cout << ff << "\n";
#endif

      for( const varatom& b : ff.r. get_lhs( ))
      {
         if( subsumes( at1, inst1, b, ff. inst ))
            return true;
      }

      if( ff.r. isdisjunctive( ))  
      {
         const auto& rhs = ff.r. getdisjunctive( ). get_rhs( ); 

         for( const varatom& b : rhs )
         {
            if( subsumes( at1, inst1, b, ff. inst ))
               return true;
         }

         return false;
      }

      return false; 
   }
}

   bool subsumes( const false_instance& f1, const false_instance& f2 )
   {
#if 0
      std::cout << "subsumption:\n";
      std::cout << f1 << "\n";
      std::cout << f2 << "\n";
#endif

      if( f1.r. isexistential( ))
      {
         if( ! f2.r. isexistential( )) 
            return false;

         const auto& ex1 = f1.r.getexistential( );
         const auto& ex2 = f2.r.getexistential( );

         if( ! subsumes( ex1. getvariable( ), ex1. getatom( ), f1. inst, 
                         ex2. getvariable( ), ex2. getatom( ), f2. inst ))
         {
            return false;
         }
   
         for( const auto& at : ex1. get_lhs( ))
         {
            if( !implies( at, f1. inst, f2 ))
               return false;
         }
         return true; 
      }
      else
      {
         const auto& disj = f1. r. getdisjunctive( ); 

         for( const varatom& a : disj. get_lhs( ))
         {
            if( !implies( a, f1.inst, f2 ))
               return false;
         }

         for( const varatom& a : disj. get_rhs( ))
         {
            if( !implies( a, f1.inst, f2 ))
               return false;
         }

         return true; 
      }
   }

   
   bool isbetterthan( const false_instance& f1, const false_instance& f2 )
   {

      size_t c1 = f1. nrchoices;
      if( c1 > 2 ) c1 = 2;

      size_t c2 = f2. nrchoices;
      if( c2 > 2 ) c2 = 2;

      // We prefer closure over deterministic over branching: 

      if( c1 < c2 ) return true;
      if( c1 > c2 ) return false;

      // Multiset comparison of premisses used:

      int c = multiset_compare( f1. assumptions, f2. assumptions );
      if( c == -1 )
         return true;
      if( c == 1 )
         return false;

      // We prefer smaller number of premisses:

      size_t s1 = f1.r.get_lhs( ).size( );
      size_t s2 = f2.r.get_lhs( ).size( );

      if( s1 < s2 ) return true;
      if( s1 > s2 ) return false;

      // Number choices.

      if( f1. nrchoices < f2. nrchoices )
         return true;
      if( f1. nrchoices > f2. nrchoices )
         return false;

      return false;
   }

}

