
// Written by Hans de Nivelle, Dec. 2014.
// Removed template parametrization in May 2015.


#include "findfalse.h"
#include "matching/basics.h"
#include "../statistics/distribution.h"
#include "../statistics/reporter.h"
#include "../statistics/timer.h"



void geometric::check_rhs( const rule& r, 
           const subst<element> & inst, 
           const interpretation& intp, 
           false_instance_set& solutions, size_t& solutioncounter )
{
#if 0
   std::cout << "Checking rhs "; 
   std::cout << r << "\n";
   std::cout << inst << "\n\n";
#endif    

   if( r. isdisjunctive( ))
   {
      const disjunctive& disj = r. getdisjunctive( );
        
      if( !intp.conclusionistrue( disj, inst ))
      {
         false_instance ff( r, 0,
                            inst, intp. getassumptions( r, inst ));
         ff. nrchoices =
                  intp. checkchoices( disj, inst, ff. assumptions );
         solutions. insert( std::move( ff ));
         ++ solutioncounter;
      }
      return;
   }

   if( r. isexistential( ))
   {
      const existential& exists = r. getexistential( );
 
      if( !intp.conclusionistrue( exists, inst ))
      {
         false_instance ff( r, 0,
                            inst, intp. getassumptions( r, inst ));
         ff. nrchoices =
                 intp. checkchoices( exists, inst, ff. assumptions );
         solutions. insert( std::move( ff ));
         ++ solutioncounter; 
      }
      return;
   }

   // Should be unreachable:

   throw std::logic_error( "forward reasoning: unknown rule type" );
}

   
void geometric::findfalse( const indexedrule& from, 
                           size_t level, 
                           const interpretation& intp,
                           subststack<element> & subst, 
                           false_instance_set& solutions,
                           size_t& solutioncounter )
{
#if 0
   std::cout << "findfalse on level " << level << ":\n";
   std::cout << from. r << "\n";
   std::cout << subst << "\n";
#endif

   if( level == from. prem. size( ))
   {
      // We have a solution: 

      check_rhs( from.r, subst.s, intp, solutions, solutioncounter );

      return; 
   }

   const varatom& atom = from. prem[ level ];

   // Equality: The variables must
   // be assigned (this is a precondition) and distinct.

   if( atom. p. isequality( ))
   {
      auto p1 = subst. clookup( atom[0] );
      auto p2 = subst. clookup( atom[1] );
      if( !p1 || !p2 )
            throw std::logic_error( "equality has unassigned variables" );

      if( *p1 != *p2 )
      {     
         findfalse( from, level+1, intp, subst, solutions, solutioncounter ); 
      }
   }
   else
   {
      std::vector< range< size_t >> occ = 
               matching::getranges( atom, subst.s, intp. atoms );

      // We enumerate common elements, and try to match:

      size_t s = subst. size( );   // For restoring.

      bool b = findcommon( occ );
      while(b)
      {
         size_t k = * ( occ[0] ++ );
         
         const groundatom& into = intp. atoms [k];

         if( ( atom.t & into.t ) == truthval::empty && 
               subst. match_arguments( atom, into ))
         {
            findfalse( from, level+1, intp, subst, 
                       solutions, solutioncounter ); 
         }

         // Restore subst. This must always be done, also when
         // matching failed.

         subst. restore(s);

         // Retrieve a next candidate matching atom: 

         b = findcommon( occ );
      }
   }

}


void geometric::findfalse( const indexedrule& from,
                           const interpretation& intp,
                           false_instance_set& solutions )
{
   subststack<element> ss;

   static statistics::distribution times( 1.0E-3, 1.0E4, 21, true );
   static unsigned int counter = 0;
   static statistics::reporter r;

   static double nextreport = 1.0;   // Next time that we will report.

   if( r. mustreport( ))
   {
      std::cout << "Report from findfalse:\n";
      std::cout << "counter = " << counter << "\n";
      r. printduration( std::cout );
      std::cout << times << "\n";
      r. wasreported( );
   }

   ++ counter;

   size_t solutioncounter = 0;

   statistics::timer t{ [ &from, &intp, &solutioncounter ]( double d ) 
      { times. insert(d); 
      if( d >= nextreport )
      {
         std::cout << "\n";
         std::cout << "-----------------------------------------\n";
         std::cout << "found hard matching instance !!!!\n";
         std::cout << "-----------------------------------------\n";
         std::cout << from.r << "\n\n";
         std::cout << intp << "\n\n";
         std::cout << "It took " << d << " seconds to complete\n";
         std::cout << "It had " << solutioncounter << " solutions\n\n\n";
         while( d >= nextreport )
            nextreport *= 2.0;
      } } };

   
   findfalse( from, 0, intp, ss, solutions, solutioncounter ); 
}


geometric::false_instance_set
geometric::findconflict( const std::vector< indexedrule > & conflicts,
                         const interpretation& intp, size_t maxsol )
{
   false_instance_set found{ maxsol };

   for( const indexedrule& c: conflicts )
   {
      findfalse( c, intp, found );
   }

   return found;
}


geometric::false_instance_set
geometric::findfalse( const rulesystem& rs, 
                      const interpretation& intp, size_t maxsol ) 
{
   false_instance_set found{ maxsol };

   for( const systemrule& r : rs. initials )
   {
      findfalse( r. ir, intp, found ); 
   }

   for( const systemrule& r : rs. learnt )
   {
      findfalse( r. ir, intp, found );
   }

   return found; 
}




