
// Written by Hans de Nivelle, June/September 2014.

#include "formula.h"
#include "../tostring.h"
#include <map>


geometric::formula::formula( std::initializer_list< varatom > lhs )
{
   for( const varatom& at : lhs )
      insert_lhs( at );
}


geometric::formula::formula( const std::vector< varatom > & lhs )
{
   for( const varatom& at : lhs )
      insert_lhs( at );
}
 

void geometric::formula::insert_lhs( varatom at )
{
   if( at. p. iscreate( ) && at.t != truthval::F ) 
   {
      throw std::runtime_error( std::string( "cannot insert " ) +
         tostring( at ) +
         " into formula, because create must have truth-value F" );
   }

   if( at. p. isequality( ) && at.t != truthval::T ) 
   {
      throw std::runtime_error( std::string( "cannot insert " ) +
         tostring( at ) +
         " into formula, because equality must have truth-value T" );
   }

   // We always merge with existing atoms, when possible:

   auto p = lhs. begin( ); 
   while( p != lhs. end( ))
   {
      if( mergewith( *p, at )) 
         p = lhs. erase( p );  
      else
         ++ p;
   }

   lhs. push_back( at );
}



void geometric::formula::print_lhs( std::ostream& out ) const 
{
   if( lhs. size( ))
   {
      for( auto p = lhs. begin( ); p != lhs. end( ); ++ p )
      {
         if( p != lhs. begin( ))
            out << ", ";
         out << *p;
      }
   }
   else
      out << "TRUE";
}



namespace geometric 
{ 
   namespace 
   { 

      // Returns number of distinct, unassigned variables:

      unsigned int nr_unassigned( 
                    const std::map< variable, bool > & assigned,
                    const varatom& at )
      {
         unsigned int u = 0;
         for( auto p = at. args_cbegin( ); p != at. args_cend( ); ++ p )
         {
            // We count unassigned, first time occurrences:

            if( assigned. find( *p ) == assigned. end( ))
            {
               bool firsttime = true;
               for( auto q = at. args_cbegin( ); q != p && firsttime; ++ q ) 
               { 
                  if( *p == *q )
                     firsttime = false; 
               }

               if( firsttime )
                  ++ u;
            }
         }
         return u;
      }


      // Preference categories are as follows: 
      // 0. Fully assigned equality or create atoms.
      // 1. Regular atoms.
      // 2. Unassigned create atoms.
      // 3. Unassigned equalities.

      unsigned int category( const std::map< variable, bool > & assigned,
                             const varatom& at )
      {
         if( at.p.isequality( ) && at.t == truthval::T )
         {
            if( nr_unassigned( assigned, at ) == 0 )
               return 0;
            else
               return 3;
         }

         if( at.p.iscreate( ) && at.t == truthval::F )
         {
            if( nr_unassigned( assigned, at ) == 0 )
               return 0;
            else
               return 2;  
         }    

         return 1; // Regular atom.
      }

 
      // If the last variable of at1 is unassigned, 
      // and occurs in at2 at a position that
      // is not last, then at1 should be assigned before at2.
      // The idea behind this rule is that functional chains of form
      // A(X), P(X,Y), P(Y,Z), P(Z,T) will be matched in the right order.

      bool prefer_chain( const std::map< variable, bool > & assigned,
                         const varatom& at1, const varatom& at2 )
      {
         if( at1. getarity( ) == 0 || at2. getarity( ) == 0 )
            return false;

         if( at1.p.iscreate( ) || at1.p.isequality( ) ||
             at2.p.iscreate( ) || at2.p.isequality( ))
         {
            return false;
         }

         variable last = at1[ at1. getarity( ) - 1 ];
         if( assigned. find( last ) != assigned. end( ))
            return false;

         for( size_t i = 0; i < at2. getarity( ) - 1; ++ i )
         {
            if( last == at2[i] )
               return true;
         }
         return false; 
      }
   }
}


void geometric::formula::sort_lhs( )
{
   if( lhs. size( ) >= 10000 )
      throw std::runtime_error( "sort_lhs( : left hand side is too big" ); 

      // This would screw up our point counting.
      // If it happens, the method needs to be reimplemented.

   std::vector< varatom > matched;
      // This vector will eventually replace passive.

   std::map< variable, bool > assigned;
      // These are the variables that occur in matched.
      // If a naive matching algorithm matches the atoms
      // in sorted, the variables in assigned obtain a value.

   while( lhs. size( ))
   {
#if 0 
      std::cout << "assigned:   ";
      for( auto p = assigned. begin( ); p != assigned. end( ); ++ p )
         std::cout << "   " << ( p -> first );
      std::cout << "\n";

      std::cout << "left hand side:   ";
      for( auto p = lhs. begin( ); p != lhs. end( ); ++ p )
         std::cout << "   " << *p;
      std::cout << "\n";
#endif

      // A vector of zeroes with the size of passive:

      std::vector< unsigned int > penalties = 
         std::vector<unsigned int> ( lhs. size( ), 0 );

      for( size_t i = 0; i < lhs. size( ); ++ i )
      {
         penalties[i] += 100 * nr_unassigned( assigned, lhs[i] );

         for( size_t j = 0; j < lhs. size( ); ++ j )
            if( i != j )
            { 
               unsigned int c1 = category( assigned, lhs [i] );
               unsigned int c2 = category( assigned, lhs [j] );
               if( c1 < c2 )
                  penalties[j] = 100000000; 
           
               if( c1 == c2 )
               {
                  if( prefer_chain( assigned, lhs[i], lhs[j] ))
                     penalties[j] += 1;
               }
            }
      }

      size_t best = 0;
      for( size_t i = 1; i < penalties. size( ); ++ i )
      {
         if( penalties[i] < penalties[ best ] )
            best = i;
      }

#if 0
      std::cout << "penalties:   "; 
      for( auto p = penalties. begin( ); p != penalties. end( ); ++ p )
         std::cout << "   " << *p;
      std::cout << "\n";

      std::cout << "best = " << best << "\n";
#endif

      // If best is not an instantiated #_{F} V atom, we will copy it:
 
      if( ! ( lhs[ best ].p.iscreate( ) &&
              lhs[ best ].t == truthval::F &&
              nr_unassigned( assigned, lhs[ best ] ) == 0 ))
      {
         // Mark variables in best as assigned:
 
         for( auto a = lhs[ best ]. args_cbegin( ); 
                   a != lhs[ best ]. args_cend( ); 
                   ++ a )
         {
            assigned [ *a ] = true;
         }

         matched. push_back( lhs[ best ] );
      }

      // Finally, we erase best from the left hand side:

      lhs. erase( lhs. begin( ) + best ); 
   }

   lhs = std::move( matched );
}


void geometric::formula::removeredundant( ) 
{
   auto p = lhs. begin( );
   while( p != lhs. end( ))
   {
      const varatom& at = *p; 
      if( at.p.iscreate( ) && at.t == truthval::F )
      {
         variable v = at[0];
         if( !isfree( v, p ) && hasothervars( v ))
            p = lhs. erase( p );
         else
            ++ p;
      }
      else
         ++ p;
   }
}


geometric::set< geometric::variable >
geometric::formula::getlhsvariables( ) const
{
   set< variable > result;

   for( const varatom& at : lhs )
   {
      for( variable v: at )
         result. insert(v);
   }

   return result;
}
 

bool geometric::formula::isfree( variable v, 
        const std::vector< varatom > :: const_iterator it ) const
{
   for( auto p = lhs. begin( ); p != lhs. end( ); ++ p )
   {
      if( p != it )
      {
         for( variable w : *p )
         {
            if( v == w ) 
               return true;
         }
      }
   }
   return false;
}


bool geometric::formula::hasothervars( variable v ) const
{
   for( const varatom& at : lhs )
   {
      for( variable w : at )
         if( v != w ) 
            return true;
   }
   return false;
}

 
void geometric::formula::makerangerestricted( )
{
   // Collect all variables occurring in atoms in the lhs that
   // need guards. We cannot insert immediately, because insertion
   // may invalidate iterators.

   std::vector< variable > eq;
   const predicate* p = nullptr;
      // Will be assigned when we find an atom that needs a guard.

   for( const varatom& at : lhs )
   {
      if( needsguard(at) )
      {
         p = &at.p;

         for( variable v : at ) 
            eq. push_back(v);
      }
   }

   // Insert won't make redundant insertions, so we just call insert
   // on all of them:

   for( variable v : eq )
   {
      insert_lhs( varatom( p -> getcreatefromsignature( ), 
                           truthval::F, { v } ));
   }

}


bool geometric::formula::needsguard( const varatom& at )
{
   return at.p. isequality( ) && at.t == truthval::T;
}


bool geometric::formula::ispassive( const varatom& at )
{
   if( at.p.isequality( ) && at.t == truthval::T )
      return true;

   if( at.p.iscreate( ) && at.t == truthval::F )
      return true;

   if( ( at.t & truthval::F ) != truthval::empty )
      return true;  

   return false; 
}


bool geometric::formula::mergewith( const varatom& at1, varatom& at2 )
{

   if( at1.p == at2.p && equal_arguments( at1, at2 ))
   {
      at2.t |= at1.t;
      return true;
   }

   if( at1.p.iscreate( ) && at1.t == truthval::F && 
       !needsguard( at2 ) && at2. contains( at1[0] ))
   {
      return true;  
   }

   if( at2.p.iscreate( ) && at2.t == truthval::F && 
       !needsguard( at1 ) && at1. contains( at2[0] ))
   {
      at2 = at1;
      return true;
   }

   // Equality is commutative. We could also merge truth values,
   // but this is enough, I think:

   if( at1.p.isequality( ) && at2.p.isequality( ) && at1.t == at2.t )
   {
      if( at1[0] == at2[1] && at1[1] == at2[0] ) 
         return true;
   } 
       
   return false; 

}



