
// Written by Hans de Nivelle, August 2014.
// In May 2015, I removed the lemma type.


#ifndef GEOMETRIC_FORMULA_INCLUDED
#define GEOMETRIC_FORMULA_INCLUDED  1

#include "../unique_number.h" 
#include "atom.h"
#include "set.h" 
#include <vector>

namespace geometric
{
   // formula is an abstract class from which the concrete formulas 
   // inherit. The concrete rules are: 
   //    disjunctive:  A formula with an unbounded number of rhs atoms.
   //    existential:  A formula with 1 existentially quantified rhs atom. 

   class formula
   {
   private: 
      unique_number< formula > number;

   protected:
      std::vector< varatom > lhs; 

      formula( ) = default;
      formula( std::initializer_list< varatom > lhs );
      formula( const std::vector< varatom > & lhs );
         // The constructors use insert_lhs to avoid redundancy. 

      formula( const formula& ) = default;
      formula( formula&& ) noexcept = default;
      formula& operator = ( const formula& ) = default;
      formula& operator = ( formula&& ) = default;

   public: 
      virtual formula* clone( ) const & = 0; 
      virtual formula* clone( ) && = 0;
 
      virtual ~formula( ) { } 

      void insert_lhs( varatom at );
         // Inserts at into the lhs of the rule. 
         // Insertion always tries to merge with existing atoms.
         // Which atoms can be merged is decided by merge_with. 

      const std::vector< varatom > & get_lhs( ) const
         { return lhs; } 

      size_t lhs_size( ) const 
         { return lhs. size( ); }
     
      virtual void print( std::ostream& ) const = 0;  

      void printnumber( std::ostream& out ) const
         { out << '#' << number; }

      void print_lhs( std::ostream& out ) const;
     
      void sort_lhs( );
         // This method sorts the passive atoms with two aims: 
         // The resulting order must be suitable for naive matching 
         // algorithms, and the rule must look good. 

      virtual void makerangerestricted( ); 
         // Make the rule range restricted by inserting suitable
         // #_{F} V predicates. A rule is range restricted if
         // every variable that is free in the rhs, or occurs 
         // in an equality in the lhs, also occurs in a non-equality
         // predicate in the lhs.  
         // Don't make assumptions about where the #_{F} predicates
         // are inserted. The #_{F} predicates are built using the
         // signature of the predicate symbol of the atom that needs the 
         // #_{F} guard. 

      void removeredundant( ); 
         // Remove redundant atoms from lhs. Currently we delete only
         // atoms #_{F} V, for which 
         //    V is not free in the formula, (apart from #_{F} V), and  
         //    V is not the only variable in the formula.  

      set<variable> getlhsvariables( ) const;
         // Gets the variable in the left hand side. If we are range
         // restricted, these are all variables.

      bool isfree( variable v ) const { return isfree( v, lhs. end( )); }
            // Return true if v is free in the formula. 
 
      virtual bool hasothervars( variable v ) const;
         // Return true if the formula contains a free variable that differs
         // from v. 

      virtual bool isfree( variable v,
                 std::vector< varatom > :: const_iterator it ) const;
         // Return true if the variable has a free occurrence in an atom
         // at another position than it.

      static bool needsguard( const varatom& at );
         // True if atom at needs guards for its variables.
         // Only positive equalities need guards. 
         // A guard is a predicate of form #_{F} V.

      static bool ispassive( const varatom& at );
         // True if atom is passive, which is the case if
         //    1. it has form X = Y,
         //    2. it has form #_{F} X,
         //    3. it has form P_{T}( ... ), with F in T.

      static bool mergewith( const varatom& at1, varatom& at2 );
         // Try to merge at1 into at2 and return true if merging succeeds. 
         // In that case, at1 can be deleted. 
         // We merge p_{T1}( ) with p_{T2}( ) if the arguments are equal.
         // We merge #_{F} X with p_{T}( ) if X occurs among the arguments.
         // We merge X = Y with Y = X.
         // If merging happens, then the following must hold: 
         //    result in conflict ==> 
         //       the two (original) premisses were in conflict. 


   };


   inline std::ostream& operator << ( std::ostream& out, const formula& f )
   {
      f. print( out );
      return out;
   }

}

#endif

