
// Written by Hans de Nivelle, April 2015.

#ifndef GEOMETRIC_INFERENCE_INCLUDED
#define GEOMETRIC_INFERENCE_INCLUDED  1

#include <chrono>
#include "age.h"

namespace geometric
{

   struct inference;

   // Infbase is the base class from which inference rules must inherit:
   // Infbase is supposed to exist only on the heap.

   struct infbase
   {
      age a;

      size_t counter;  // Reference counter.

      infbase( ) : counter{1} 
         { }

      virtual ~infbase( ) { };

      virtual infbase* clone( ) const & = 0;
      virtual infbase* clone( ) && = 0;
         // Because without reference, clone( ) const would bind also
         // to an Rvalue reference. 

      virtual bool iscorrect( ) const 
      {
         std::cerr << "checking this inference was not implemented\n";
         return false;
      }
         // Checks correctness of the inference. (Only this step.)

      // Access to the parents of this inference:

      virtual size_t nrparents( ) const = 0;
      virtual const inference& operator [] ( size_t i ) const = 0;
         // Get the i-th parent.  

      virtual void print( std::ostream& ) const = 0;
         // Printing recurses through the proof tree, but stops 
         // when it encounters a rule. 
   };


   class inference
   {
      infbase* inf;

   public: 
      inference( const infbase& b ) 
         : inf{ b. clone( ) }
      { }

      inference( infbase&& b )
         : inf{ std::move(b). clone( ) }
      { }

      inference( const inference& other ) noexcept 
         : inf{ other. inf }
      {
         ++ ( inf -> counter );
      }

      void operator = ( inference&& other )
         { std::swap( inf, other. inf ); }
 
      void operator = ( const inference& other )
         { *this = inference{ other }; } 

      ~inference( )
      {
         -- ( inf -> counter );
         if( inf -> counter == 0 )
            delete inf;
      }
    
      bool iscorrect( ) const { return inf -> iscorrect( ); }

      size_t nrparents( ) const { return inf -> nrparents( ); }
      const inference& operator [] ( size_t i ) const 
         { return (*inf)[i]; }

      void print( std::ostream& out ) const
         { inf -> print( out ); } 

      const infbase* getcontents( ) const { return inf; }

   };


   inline std::ostream& operator << ( std::ostream& out, 
                                      const inference& inf )
   {
      inf. print( out );
      return out;
   }

 
   // Inference rule for rules that are part of the code:

   struct partofcode : public infbase
   {
      partofcode* clone( ) const& override
         { return new partofcode( *this ); }

      partofcode* clone( ) && override
         { return new partofcode( std::move( *this )); }

      bool iscorrect( ) const override { return true; }
         // We have no choice than to trust the person who typed the code. 

      size_t nrparents( ) const override 
         { return 0; } 

      const inference& operator [] ( size_t ) const override
      {
         throw std::runtime_error( "inference partofcode has no parents" );
      }

      void print( std::ostream& out ) const override
         { out << "partofcode"; }

   };

  
   // Inference rule for rules that are not sound:

   struct unsound : public infbase
   {
      unsound* clone( ) const& override
         { return new unsound( *this ); }
   
      unsound* clone( ) && override
         { return new unsound( std::move( *this )); }

      bool iscorrect( ) const override { return false; }
         // We have no choice than to trust the person who typed the code.

      size_t nrparents( ) const override
         { return 0; }

      const inference& operator [] ( size_t ) const override
      {
         throw std::runtime_error( "inference unsound has no parents" );
      }

      void print( std::ostream& out ) const override
         { out << "unsound"; }

   };


   // Inference rule for rules that originate from input: 

   struct input : public infbase
   {
      input* clone( ) const& override
         { return new input( *this ); }

      input* clone( ) && override
         { return new input( std::move( *this )); }

      bool iscorrect( ) const override { return false; }
         // We have no choice than to trust the person who typed the code.

      size_t nrparents( ) const override
         { return 0; }

      const inference& operator [] ( size_t ) const override
      {
         throw std::runtime_error( "inference input has no parents" );
      }

      void print( std::ostream& out ) const override
         { out << "input"; }

   };

}


#endif

