
// Written by Hans de Nivelle, February 2015.
// Revised in April 2015. 
// An interpretation I has form ( E, M, G, A, J ), 
// where E is a set of elements, 
// M is a set of atoms,
// G is a set of guessers,
// A is an assumption function, 
// J is a set of justifications. 
// Actually all the sets are stacks in reality. 

#ifndef GEOMETRIC_INTERPRETATION_INCLUDED
#define GEOMETRIC_INTERPRETATION_INCLUDED   1

#include <vector>
#include <iostream>

#include "age.h"
#include "permutationgenerator.h"
#include "false_instance.h"
#include "rulesystem.h"

namespace geometric
{

   struct interpretation
   {

      struct guesser 
      {
         size_t at_size;       // Size of atoms at moment of creation. 
         size_t deriv_size;    // Size of derivations at moment of creation.

         std::vector<size_t> choices; 
            // Vector of possible choices (Called S+ in the paper),
            // permuted in some random way.  

         size_t j;  
            // Choice that is currently being extended. 

         unsigned int closure_counter; 
            // Number in closure counter of the interpretation 
            // when this guess was guessed.

         age time;
            // Time on which this guess was made. 

         std::vector< indexedrule > closing;   
            // For every choice < j we have a closing lemma.
            // There is no point in keeping instantiations, because
            // we try to reuse closing lemmas between different
            // branches.  

         guesser( size_t at_size, 
                  size_t deriv_size, 
                  std::vector<size_t> && choices, 
                  unsigned int closure_counter )
            : at_size{ at_size }, 
              deriv_size{ deriv_size }, 
              choices{ std::move(choices) }, 
              j{0},
              closure_counter{ closure_counter }
         { } 

         void addclosinglemma( const rule& r )
            { closing. push_back(r); } 
      };

      std::vector< guesser > guesses;

      std::vector< element > elements;
      index::standardsimple< element > atoms;  
      std::vector< set< size_t >> assumptions; 
         // Always atoms. size( ) == assumptions. size( ). 
         // Initial atoms have assumptions[i] = { }. 

      std::vector< false_instance > derivations; 
         // Initial atoms have no derivation. 
         // Deterministically derived atoms have a derivation with one choice.
         // Guesses have derivations with more than one choice. 
 
      unsigned int closure_counter;


      interpretation( const std::vector< groundatom > & start );

      size_t findcreate( element e ) const;
         // Find the create atom for e.

      size_t find( const varatom& at, const subst<element> & subst ) const;
         // Return an index i, s.t. atoms[i] equals subst(at), or is
         // in conflict with subst(at). Returns atoms. size( ) if no
         // such atom exists. Substitution subst must instantiate at fully. 

      size_t find( variable v, const varatom& at,
                   const subst<element> & subst ) const;
         // Return an index i, s.t. there exists an instantiation v:=e, 
         // for which subst+{v:=e}(at) equals atoms[i].

      size_t findconflict( const varatom& at, 
                           const subst<element> & subst ) const;
      size_t findconflict( const varatom& at,
                           const subst<element> & subst,
                           variable v, element e ) const; 

         // Return index i, st.t. atoms[i] is in conflict with at
         // instantiated by subst, or atoms. size( ) if no such atom exists.
         // If v occurs in subst, then assignment { v:=e } overrides subst.

      size_t conflictlevel( const disjunctive& disj, 
                            const subst<element> & subst ) const;
         // Return the smallest number s, s.t. at is in conflict with 
         // the atoms in our interpretation with index < s. Returning 
         // atoms. size( ) means that atoms was in conflict, and that
         // atoms [ atoms. size( ) - 1 ] was used. 
         // A return value > atoms. size( ) means that lem is not in
         // conflict. 


      bool conclusionistrue( const disjunctive& disj, 
                             const subst<element> & subst ) const;
      bool conclusionistrue( const existential& exists,
                             const subst<element> & subst ) const;
         // True if the conclusion of the rule is true.

      bool conclusionconflicts( const disjunctive& disj,
                                const subst<element> & subst,
                                size_t j ) const;
      bool conclusionconflicts( const existential& exists,
                                const subst<element> & subst,
                                size_t j ) const;
         // True if the conclusion is in conflict. 
         // For example, conclusion is P_T( ) and P_E( ) occurs in the
         // interpretation. 
 
      void extend( const disjunctive& disj, const subst<element> & subst,
                   size_t j, const set< size_t > & a );
         // Extend interpretation by the j-th active atom of 
         // disj. Nothing is checked.

      void extend( const existential& exists, const subst<element> & subst,
                   size_t j, const set< size_t > & a );
         // Extend interpretation by conclusion of exists, instantiated
         // by j-th element. Nothing is checked.

      void pop_atom( );
         // Remove the last atom. If it is a predicate of form
         // #_t e, we also remove e. 

      guesser& addguesser( std::vector< size_t > && choices );
        // Add a guesser (backtracker). No atoms or elements are added.

      void backtrackassumptions( );
         // Remove the assumptions that were introduced after the last guess. 
         // The guess itself and its derivation are not removed.
         // Deterministic extensions are not properly handled. 
         // They should be removed first.

      void removedeterministic( );
         // Remove a deterministic extension and its derivation.
         // When the rule was existential, two atoms are removed.

      void pop_guesser( );
         // Remove the last guesser and its derivation. 
         // If there are aassumptions deterministically derived atoms , they
         // have to be withdrawn first. 

      guesser& lastguesser( ) { return guesses. back( ); }
      const guesser& lastguesser( ) const { return guesses. back( ); }

      element lastelement( ) const { return elements. back( ); }

      // A guesser has one derivation, independent of its number
      // of assumed atoms. Similarly, a deterministic extension always has
      // one derivation. 

      false_instance& addderivation( const false_instance& f ) 
         { derivations. push_back(f); return derivations. back( ); }
      false_instance& addderivation( false_instance&& f )
         { derivations. push_back(std::move(f)); return derivations. back( ); }

      const false_instance& lastderivation( ) const 
         { return derivations. back( ); }
      const false_instance& lastguessderivation( ) const;
         // The false_instance with which the last guess was made.

      void pop_derivation( ) { derivations. pop_back( ); }

      size_t nrguesses( ) const { return guesses. size( ); }
      size_t atoms_size( ) const { return atoms. size( ); }
      size_t nrelements( ) const { return elements. size( ); }
      size_t nrderivations( ) const { return derivations. size( ); } 

      set< size_t > getassumptions( const rule& r, 
                                    const subst<element> & inst ) const;
         // Collect the assumptions of the atoms that the passive atoms 
         // of rule r are in conflict with. 

      size_t checkchoices( const disjunctive& disj,
                           const subst<element> & subst,
                           set< size_t > & a ) const; 
      size_t checkchoices( const existential& exists,
                           const subst<element> & subst,
                           set< size_t > & a ) const;
         // Check each possible extension for immediate conflict. 
         // For those extensions that cause immediate conflict, 
         // we add the assumptions used by this conflict to a. 
   };

   std::ostream& operator << ( std::ostream& out, const interpretation& i );
   std::ostream& operator << ( std::ostream& out, 
                               const interpretation::guesser& );


   ///////////////////////////////////////////////////////
   //  A few inline functions:


   inline bool 
   interpretation::conclusionistrue( const disjunctive& disj,
                                     const subst<element> & subst ) const
   {
      for( const varatom& at : disj. get_rhs( ))
      {
         if( find( at, subst ) != atoms. size( ))
         {
            return true;
         }
      }

      return false;
   }


   inline bool 
   interpretation::conclusionistrue( const existential& exists,
                                     const subst<element> & subst ) const
   {
      return find( exists. getvariable( ), exists. getatom( ), subst )
                              != atoms. size( );
   }


   inline bool
   interpretation::conclusionconflicts( const disjunctive& disj,
                                        const subst<element> & subst,
                                        size_t j ) const
   {
      return findconflict( disj[j], subst ) != atoms. size( );
   }

   inline bool
   interpretation::conclusionconflicts( const existential& ex,
                                        const subst<element> & subst,
                                        size_t j ) const
   {
      if( j < elements. size( ))
         return findconflict( ex. getatom( ), subst, 
                    ex.getvariable( ), elements[j] ) != atoms. size( );

      return false;
         // Extending never conflicts, because the fresh variable 
         // must occur in the atom. 
   }


   inline size_t 
   interpretation::checkchoices( const disjunctive& disj,
                                 const subst<element> & subst,
                                 set< size_t > & a ) const
   {
      size_t choices = 0;
      for( size_t j = 0; j != disj. rhs_size( ); ++ j )
      {
         size_t s = findconflict( disj[j], subst );

         if( s < atoms. size( ))
            a |= assumptions[s];
         else
            ++ choices; 
      }
      return choices;
   }

   inline size_t 
   interpretation::checkchoices( const existential& exists,
                                 const subst<element> & subst,
                                 set< size_t > & a ) const
   {
      size_t choices = 0;
      for( size_t j = 0; j < nrelements( ); ++ j )
      {
         size_t s = findconflict( exists. getatom( ), subst, 
                                  exists. getvariable( ), elements [j] );

         if( s < atoms. size( )) 
            a |= assumptions[s]; 
         else
            ++ choices;
      }

      return choices + 1;  
         // Because adding a new element is always possible.
   }
 
}


#endif

