
// Written by Hans de Nivelle, January 2015.
// Changed the proof logging mechanism in April 17th.

#ifndef GEOMETRIC_RULE_INCLUDED
#define GEOMETRIC_RULE_INCLUDED  1

#include <chrono>

#include "age.h"
#include "formula.h"
#include "disjunctive.h"
#include "existential.h"
#include "inference.h"

namespace geometric
{

   class rule
   {
      formula* form;
      size_t* counter;
         // Reference counter for rul.

      inference inf;

   public: 
      rule( const formula& f, const inference& inf = unsound( ))
         : form{ f. clone( ) }, counter{ new size_t{1}}, inf{ inf } 
      { }

      rule( formula&& f, const inference& inf = unsound( ))
         : form{ std::move(f). clone( ) }, 
           counter{ new size_t{1}}, 
           inf{ inf }
      { }

      rule( const rule& r ) noexcept 
         : form{ r.form }, counter{ r. counter }, inf{ r.inf }
      {
         ++ ( *counter ); 
      }

      void operator = ( rule&& r )
      {
         std::swap( form, r. form );
         std::swap( counter, r. counter ); 
         std::swap( inf, r. inf );
      }

      void operator = ( const rule& r )
      {
         *this = rule{r}; 
      }

      ~rule( ) 
      {
         if( -- (*counter) == 0 ) 
         {
            delete form; 
            delete counter;
         }
      }

      const inference& getinference( ) const { return inf; }

      const std::vector<varatom> & get_lhs( ) const 
         { return form -> get_lhs( ); }

      set<variable> getlhsvariables( ) const 
         { return form -> getlhsvariables( ); }

      void sort_lhs( ) 
         { form -> sort_lhs( ); }

      void set_rhs( )
         { disjunctive& disj = dynamic_cast< disjunctive& > ( *form );
           disj. set_rhs( ); 
         }

      void makerangerestricted( )
         { form -> makerangerestricted( ); }

      void removeredundant( )
         { form -> removeredundant( ); }

      bool isdisjunctive( ) const 
         { return dynamic_cast< disjunctive* > ( form ); }
      bool isexistential( ) const
         { return dynamic_cast< existential* > ( form ); } 

      const disjunctive& getdisjunctive( ) const
         { return dynamic_cast< disjunctive& > ( *form ); }
      const existential& getexistential( ) const
         { return dynamic_cast< existential& > ( *form ); }

      void printnumber( std::ostream& out ) const
      {
         form -> printnumber( out ); 
      } 
   
      friend std::ostream& operator << ( std::ostream& , const rule& ); 

   };

   inline std::ostream& operator << ( std::ostream& out, const rule& r ) 
   {
      r.form -> printnumber( out ); out << ": ";
      out << r.inf; 
      out << ", references = " << *(r.counter);
      out << ", size of lhs = " << r.form -> lhs_size( ) << ":\n";
      out << "   "; 
      r.form -> print( out );
      return out;
   }
}

#endif


