
// Written by Hans de Nivelle, February 2015.

#include "rulesystem.h"
#include "subststack.h"
#include "subsumption.h"
#include "deduction.h"


std::ostream& geometric::operator << ( std::ostream& out,
                                       const indexedrule& ind )
{
   out << ind.r << "\n\n";
   out << ind. prem << "\n";
   out << ind. vars << "\n";
   return out;
}


std::ostream& geometric::operator << ( std::ostream& out,
                                       const systemrule& sr )
{
   out << sr.ir.r << "\n";      // Indexedrule without the index.
   out << "      (used " << sr. nruses << " times, uses = ";
   out << sr. uses << ")\n";
   return out;
}


geometric::rulesystem::rulesystem( const std::string& name,
      const std::vector< rule > & initials,
      const std::vector< rule > & simplifiers )
   : name{name}
{
   for( const rule& r : initials )
      addinitial(r);  

   for( const rule& r : simplifiers )
      ( this -> simplifiers ). push_back(indexedrule(r)); 
}


geometric::rulesystem::rulesystem( const std::string& name,
      std::initializer_list<rule> initials,
      std::initializer_list<rule> simplifiers )
   : name{name}
{
   for( const rule& r : initials )
      addinitial(r); 

   for( const rule& r : simplifiers )
      ( this -> simplifiers ). push_back(indexedrule(r)); 
}


void geometric::rulesystem::addinitial( rule r )
{
   if( r. isdisjunctive( ))
      r. set_rhs( );
   r. makerangerestricted( );
   r. sort_lhs( );

   initials. push_back( indexedrule(r) ); 
}

   
void geometric::rulesystem::addlearnt( std::list<false_instance> added )
{
   
   if( SHOWSIMPLIFICATION )
   {
      std::cout << "Starting simplification:\n"; 
      for( const auto& p : added )
         std::cout << p << "\n\n";
      std::cout << "\n";
   }
   else
   {
      std::cout << "Considering for learning:\n";  
      for( const auto& p : added )
         std::cout << p.r << "\n";
      std::cout << "\n";
   }

   for( const auto& p : added )
   {
      if( ! p. r. isdisjunctive( ))
         throw std::runtime_error( "only disjunctive formulas can be learnt" );
   }

   while( added. size( ))
   {
      false_instance& first = added. front( ); 
      
      index::standardsimple< variable > ind;

      for( const varatom& at : first. r. get_lhs( ))
         ind. insert( at );
      for( const varatom& at : first. r. getdisjunctive( ). get_rhs( ))
         ind. insert( at );

      set< variable > vars = first. r. getlhsvariables( );

      std::vector< subsumption::skippedequalities > solutions; 
         // We will build a vector of skippedequalities. Each
         // skippedequalities contains a vector of skipped equalities.
         // If one is empty, we found a full subsumption, otherwise,
         // we have a partial subsumption, and we can merge in first. 

      subststack< variable > subst;
      std::vector< std::pair< variable, variable >> skipped;
      bool stop = false;
 
      for( auto p = simplifiers. begin( ); 
                p != simplifiers.end( ) && !stop;  
                ++ p )
      {
         const disjunctive& disj = p -> ir.r. getdisjunctive( );
         subsumption::subsumes( disj, 0, 
                                ind, vars, subst, 
                                skipped, solutions, stop );
      }  

      for( auto p = learnt. begin( );
                p != learnt.end( ) && !stop;
                ++ p )
      {
         const disjunctive& disj = p -> ir.r. getdisjunctive( );
         if( disj. lhs_size( ) + disj. rhs_size( ) <= maxsimpl )
         {
            subsumption::subsumes( disj, 0, 
                                   ind, vars, subst,
                                   skipped, solutions, stop );
         }
      }

      if( solutions. size( ))
      {
         std::list< false_instance > replacement;
         for( subsumption::equality eq : solutions. front( ). skipped )
         {
            const element* e1 = first. inst. clookup( eq.v1 );
            const element* e2 = first. inst. clookup( eq.v2 ); 
  
            // One can either process all equalities, or only the 
            // true ones.

            if( *e1 == *e2 || true )
            {
               matched<disjunctive> disj = 
                  { first.r.getdisjunctive( ), first.inst };
               merging m = merging( eq.v1, eq.v2 );
               disj = merge( disj, { m } );

               replacement. push_back(
                  false_instance( 
                     rule{ std::move( disj.r ), 
                           mergings( userule( first.r ), { m } )},
                     0,
                     disj. inst, first. assumptions ));
            }
         } 
     
         added. pop_front( ); 
         for( auto& repl : replacement )
         {
            if( SHOWSIMPLIFICATION )
            {
               std::cout << "replaced by " << repl. r << "\n";
            }
            added. push_back( std::move( repl )); 
         }
      }
      else
      {
         // No solutions found means that first will be added to
         // the rule system. Insertions are always shown. 

         disjunctive disj = first.r.getdisjunctive( );
         disj. sort_lhs( ); 
         disj = normalize( disj );
         rule ins = rule( std::move(disj), first.r. getinference( ));

         std::cout << "Inserting into learnt:\n";
         std::cout << ins << "\n\n";

         learnt. push_back( systemrule( indexedrule( ins ))); 

         added. pop_front( ); 
      }
   }
}


std::ostream& 
geometric::operator << ( std::ostream& out, const rulesystem& rs )
{
   out << "RuleSystem " << rs. name << ":\n\n";
   out << "Initial Rules:\n";
   for( const auto& r : rs. initials )
      out << r << "\n";
   out << "number of initial rules = " << rs. initials. size( ) << "\n\n";

   out << "Simplifiers:\n";
   for( const auto& r : rs. simplifiers )
      out << r << "\n";
   out << "number of simplifiers = " << rs. simplifiers. size( ) << "\n\n";

   out << "Learnt:\n";
   for( const auto& r : rs. learnt )
      out << r << "\n";
   out << "number of learnt formulas = " << rs. learnt. size( ) << "\n\n";

   return out;
}


