
// Written by Hans de Nivelle, January 2015.
// Updated with indexing February 2015.

#ifndef GEOMETRIC_RULESYSTEM_INCLUDED
#define GEOMETRIC_RULESYSTEM_INCLUDED   1

#include <string>
#include <list> 
#include "rule.h"
#include "timepoints.h" 
#include "subst.h" 
#include "false_instance.h"
#include "index/simple.h"


namespace geometric
{
 
   // An indexedrule is a rule that can be matched. It consists of
   // the rule combined with an index on the atoms in the left hand side, 
   // and the set of variables. Rule must be range restricted. 

   struct indexedrule 
   {
      rule r;
      index::standardsimple< variable > prem;
      set<variable> vars;

      indexedrule( const rule& r )
         : r{r},
           vars{ r. getlhsvariables( ) } 
      {
         for( const varatom& at : r. get_lhs( ))
            prem. insert( at );
      }
   };

   std::ostream& operator << ( std::ostream& out, const indexedrule& ind );
   

   struct systemrule 
   {
      indexedrule ir;
      unsigned int nruses;
      timepoints uses;

      systemrule( const indexedrule& ir )
         : ir{ ir },
           nruses{ 0 }  
      { }

      systemrule( indexedrule&& ir )  
         : ir{ std::move(ir) },
           nruses{ 0 }
      { }

   };
 

   std::ostream& operator << ( std::ostream& out, const systemrule& sr );


   struct rulesystem
   {
      std::string name;

      std::vector<systemrule> initials;     // Rules of both possible types. 
      std::vector<systemrule> simplifiers;  // Disjunctive only. 
      std::vector<systemrule> learnt;       // Disjunctive only. 

      // Constructors sort the initial rules, and make the range
      // restricted. For this they need the signature.

      rulesystem( const std::string& name, 
          const std::vector<rule> & initials = std::vector<rule>( ),
          const std::vector<rule> & simplifiers = std::vector<rule>( ));

      rulesystem( const std::string& name,
          std::initializer_list<rule> initials,
          std::initializer_list<rule> simplifiers );

      void sort( );
         // Sort disjunctive formulas by length, or by some other 
         // complexity measure. 

      void restrict( size_t s );
         // Keep only the first s learnt rules. The learnt rules should 
         // be sorted first. 

      void addinitial( rule r );
         // 1. If the rule is disjunctive, then decide which atoms should 
         //    go to the rhs.
         // 2. Make the rule range restricted.
         // 3. Sort the premisses.
         // 4. Insert the rule. 

      void addlearnt( std::list< false_instance > );
         // Add learnt lemmas to learnt, after they have been simplified and
         // checked for subsumption. We need the full false instances, 
         // because some of the simplifications may look at the matchings.
         // The matching should make the lemma false. 

  
      static constexpr bool SHOWSIMPLIFICATION = false;
         // Show the simplification process.
         // Learnt formulas are always printed.

      static constexpr size_t maxsimpl = 7;
         // Maximum size of a disjunctive formula, so that it
         // will be used in simplification. 
   };


   std::ostream& operator << ( std::ostream& out, const rulesystem& rs );

}

#endif


