
// Written by Hans de Nivelle, Jan 2015.
// Revised in April 2015. 

#include "searchstate.h"
#include "findfalse.h"
#include "../statistics/distribution.h"
#include "../statistics/reporter.h"
#include "../statistics/timer.h" 


namespace geometric
{

   std::ostream& operator << ( std::ostream& out, const searchstate& ss )
   {
      out << "Searchstate( " << ss. starttime << " ):\n\n";

      out << "Permutation Generator:\n";
      out << "   " << ss. generator << "\n";
      out << "\n";

      out << ss.rs << "\n";

      out << ss. intp;

      return out;
   }


   inline void searchstate::justderived( matched< disjunctive > & disj )
   {
      disj.r. sort_lhs( );
      disj.r. removeredundant( );
   }


   void 
   searchstate::best_disjunctive_resolvents( 
            const matched< disjunctive > & disj,
            const inference& inf,
            const std::vector< size_t > & extensions,
            std::vector< false_instance > & chosen,
            const std::vector< false_instance_set > & choices,
            false_instance_set& best ) const 
   {
      size_t level = chosen. size( );

      if( level == choices. size( ))
      {
         // We have a complete set of choices, we construct a resolvent:

         matched< disjunctive > res = disj_resolve( disj, extensions, chosen );
         justderived( res );

         // In order to construct an inference, we need the side parents. 
         // They can be easily extracted from chosen: 

         std::vector< inference > parents = { inf }; 
         for( const false_instance& ff : chosen )
            parents. push_back( ff.r. getinference( ));

         false_instance ff(
            rule( res. r, disjres( std::move( parents ))), 0, 
                  res. inst,
            intp. getassumptions( res.r, res.inst ) );
         ff. nrchoices = 
            intp. checkchoices( res.r, res.inst, ff. assumptions );
         best. insert( std::move( ff ));
      }
      else
      {
         // We choose from choices, and continue: 

         for( const false_instance& ff : choices[ level ] )
         {
            chosen. push_back( ff );
            best_disjunctive_resolvents( disj, inf, extensions, 
               chosen, choices, best );
            chosen. pop_back( );
         }
      }
   }


   // Unwind deterministic additions until we either arrive at a 
   // guess, or have only initial atoms left.

   void searchstate::conflict( false_instance_set& confl )
   {

      if( outputlevel >= 5 )
      {
         std::cout << "Conflict:\n";
         std::cout << "---------\n";
         std::cout << confl << "\n"; 
         std::cout << intp << "\n";
      }

      while( intp. nrderivations( ) && 
             intp. lastderivation( ). nrchoices == 1 ) 
      {
         const false_instance& last = intp. lastderivation( );

         // We look for lemmas that use the last atom:

         for( false_instance& ff : confl )
         {
            size_t level = 
                      intp. conflictlevel( ff.r.getdisjunctive( ), ff. inst );
            if( level > intp. atoms_size( ))
            {
               std::cout << intp << "\n";
               std::cout << ff << "\n"; 
               throw std::runtime_error( "lemma is not a conflict lemma" ); 
            }

            if( level == intp. atoms_size( ))
            {
               if( last.r.isdisjunctive( ))
               {
                  // We determine which j caused the extension:

                  size_t j = 0;
                  while( intp. conclusionconflicts( 
                                   last.r.getdisjunctive( ), last.inst, j ))
                  {
                     ++ j;
                  }

                  matched<disjunctive> res = disj_resolve(
                     { last. r. getdisjunctive( ), last. inst }, 
                     { j }, { ff } );

                  justderived( res );

                  ff = false_instance(
                        rule( res.r, 
                              hornchain( userule( last.r ), userule( ff.r ) )),
                        0, res. inst,
                        intp. getassumptions( res.r, res. inst ));

               }
               else
               {
                  element skolem_const = intp. lastelement( );

                  matched<disjunctive> res = ex_resolve( 
                     { last. r. getexistential( ), last. inst }, 
                     skolem_const, ff );

                  justderived( res ); 

                  ff = false_instance( 
                        rule( res.r, 
                              hornchain( userule( last.r ), userule( ff.r ) )), 
                        0, res.inst,
                        intp. getassumptions( res.r, res. inst ));

               }
            } 
         }

         intp. removedeterministic( );  
      }

      if( outputlevel >= 5 )
      {
         std::cout << "Exit conflict unwind:\n";
         std::cout << confl << "\n";
         std::cout << intp << "\n";
      }
   }


   void searchstate::filterderived( false_instance_set& derived,
                                    false_instance_set& notused )
   {
      if( outputlevel >= 5 )
      {
         std::cout << "\n";
         std::cout << "Filterderived:\n";
         std::cout << "-------------\n";
         std::cout << derived << "\n";
      }

      if( derived. size( ) == 0 )
         throw std::runtime_error( "derived should not be empty" );

      size_t nrchoices = derived. front( ). nrchoices;

      for( const auto& ff : derived )
      {
         if( ff. nrchoices < nrchoices )
            nrchoices = ff. nrchoices;
      }

      if( outputlevel >= 5 )
         std::cout << "nrchoices = " << nrchoices << "\n";

      derived. erase( [ nrchoices ]( const false_instance& f )
                                { return f. nrchoices > nrchoices; }, notused );

      // IF THERE IS MORE THAN ONE, DELETE THE OTHERS?

      if( outputlevel >= 5 )
      {
         std::cout << "Exiting filterderived:\n";
         std::cout << derived << "\n";
      }
   }

 
   void searchstate::learn( false_instance_set& applied, 
                            false_instance_set& unapplied ) 
   {
   
      if( outputlevel >= 5 )
      {
         std::cout << "\n";
         std::cout << "Considering for learning:\n";
         std::cout << "-------------------------\n";
      }

      for( const auto& a : applied )
      {
         rs. addlearnt( { a } );
      }

      for( const auto& u : unapplied )
      {
         if( u.r. isdisjunctive( ))
         {
            const disjunctive& disj = u.r.getdisjunctive( );
            if( disj. rhs_size( ) < 2 && disj. get_lhs( ). size( ) < 8 )
               rs. addlearnt( {u} ); 
         }
      }
   }


   false_instance_set 
   searchstate::resolve( const matched< existential > & exists,
                         const inference& inf, 
                         element skolem_const, 
                         false_instance_set& confl )
   {

      false_instance_set result{ maxfalse }; 

      for( const auto& ff : confl ) 
      {
         matched< disjunctive > 
         disj = ex_resolve( exists, skolem_const, ff );

         justderived( disj );

         auto add = false_instance( 
                  rule( disj.r, existsres( inf, userule( ff. r ))), 0,
                        disj. inst, 
                        intp. getassumptions( disj.r, disj.inst ));

         // If some of the conclusions cause immediate conflict,
         // we do not count them as choices, and we add the assumptions.

         add. nrchoices =
            intp. checkchoices( disj.r, disj.inst, add. assumptions );
         
         result. insert( std::move( add ));
      }

      return result; 
   }


   false_instance_set
   searchstate::resolve( const matched< disjunctive > & disj, 
                         const inference& inf,
                         const std::vector< indexedrule > & conflictset )
   {

      if( outputlevel >= 8 )
      {
         std::cout << "Preparing for inference with:\n" << disj << "\n";
         std::cout << "inference " << inf << "\n";
      }
 
      std::vector< size_t > extensions; 
      std::vector< false_instance_set > conflicts;
      std::vector< size_t > nolemma; 

      // For each active atom j of disj, there are three possibilities:
      // 1. Immediate conflict. We ignore j.
      // 2. No immediate conflict and we find conflict lemmas.
      //    j is appended to extensions, and we append the conflict lemmas to 
      //    conflicts.
      // 3. No immediate extensions, but no conflict lemmas. j goes into
      //    nolemma. 

      auto aa = intp. lastderivation( ). assumptions;
      aa. insert( intp. nrguesses( ) - 1 );

      for( size_t j = 0; j < disj.r.rhs_size( ); ++ j )
      {
         if( !intp.conclusionconflicts( disj.r, disj. inst, j ))
         {
            intp. extend( disj.r, disj.inst, j, aa );

            false_instance_set closing =  
                       findconflict( conflictset, intp, maxconflict );

            if( closing. size( ) && closing. back( ). nrchoices != 0 )
               throw std::runtime_error( "there must be only closing lemmas" );

            intp. backtrackassumptions( );

            if( closing. size( ) == 0 )
            {
               nolemma. push_back(j);
            } 
            else
            {
               extensions. push_back(j); 
               conflicts. push_back( std::move( closing ));
            }
         }
      }

      if( outputlevel >= 8 )
      {
         std::cout << "Conflict Lemmas:\n";
         for( size_t i = 0; i < conflicts. size( ); ++ i )
         {
            std::cout << "------------------------------------------------\n";
            std::cout << "extending by " << extensions[i] << "  =>  ";
            std::cout << disj.r[ extensions[i] ] << "\n";
            std::cout << conflicts[i] << "\n";
         }

         if( nolemma. size( ))
         {
            std::cout << "nolemmas: { ";
            for( size_t i = 0; i < nolemma. size( ); ++ i )
            {
               if(i) std::cout << ", ";
               std::cout << nolemma[i];
            }
            std::cout << " }\n";
         }

         std::cout << "--------------------------------------------------\n";
      }

      if( nolemma. size( ) == 0 )
      {
         conflicts. pop_back( );
         extensions. pop_back( );
         nolemma. push_back( disj.r.rhs_size( ) - 1 );  
      }

      for( size_t nol : nolemma )
         extensions. push_back( nol );

      std::vector< false_instance > chosen;
         // Used for backtracking in construct_disjunctive_resolvents.

      false_instance_set best{ maxfalse };  
      best_disjunctive_resolvents( disj, inf, extensions, chosen, 
                                   conflicts, best ); 
      return best; 
   }

 
   void searchstate::try_restart( false_instance_set& derived, 
                                  false_instance_set& notused ) 
   {

      if( outputlevel >= 5 ) 
      {
         std::cout << "Entering try_restart:\n"; 
         std::cout << "---------------------\n";

         std::cout << intp << "\n";
         std::cout << derived << "\n\n";
      }

      if( derived. size( ) == 0 )
         throw std::runtime_error( "try_restart: derived is empty" );  

      // We keep only those formulas that are at lowest level:

      size_t level = intp. nrguesses( );
      for( const auto& ff : derived )
      {
         size_t l = ff. level( );
         if( l < level )
            level = l;
      }

      if( outputlevel >= 7 )
         std::cout << "level = " << level << "\n";
 
      derived. erase( [level]( const false_instance& f )
                                  { return f. level( ) > level; }, notused );

      while( intp. nrguesses( ) > level ) 
      {
         // Remove the last guess, starting with its deterministic 
         // extensions: 

         while( intp. lastderivation( ). nrchoices == 1 )
            intp. removedeterministic( );

         intp. backtrackassumptions( );
         intp. pop_guesser( );
      }

      if( outputlevel >= 5 )
      {
         std::cout << "Exiting try restart:\n\n";
         std::cout << intp << "\n";
         std::cout << derived << "\n\n";
      }
   }


   void searchstate::recheck( false_instance_set& set ) const
   {

#if 0
      std::cout << "Entering recheck:\n";
      std::cout << set << "\n";
#endif 

      false_instance_set::iterator it = set. begin( );
      while( it != set. end( ))
      {
         bool keep = false;

         if( it -> r. isdisjunctive( ))
         {
            const disjunctive& disj = it -> r. getdisjunctive( );
            if( !intp. conclusionistrue( disj, it -> inst ))
            {
               it -> nrchoices = 
                         intp. checkchoices( disj, 
                                             it -> inst, it -> assumptions );
               keep = true; 
            }
         }

         if( it -> r. isexistential( )) 
         {
            const existential& exists = it -> r. getexistential( );
            if( !intp. conclusionistrue( exists, it -> inst ))
            { 
               it -> nrchoices = 
                         intp. checkchoices( exists, it -> inst, 
                                             it -> assumptions ); 
               keep = true;
            } 
         }

         if( keep )
            ++ it;
         else
            it = set. erase( it );
      }

#if 0
      std::cout << "Exiting recheck:\n";
      std::cout << set << "\n\n";
#endif

   }


   bool searchstate::search( unsigned int maxtime )
   {
      starttime. restart( ); 

      statistics::reporter r;

   findfalse:
      if( r. mustreport( ))
      {
         std::cout << "------------------------------------------------\n";
         std::cout << "Report from search:\n";   
         r. printduration( std::cout ); 
         std::cout << "\n\n";
         std::cout << intp << "\n\n";
         std::cout << rs << "\n";
         std::cout << "------------------------------------------------\n";
         r. wasreported( );
      }

      if( outputlevel >= 9 )
      {
         std::cout << "---------------------------\n";
         std::cout << "findfalse:\n";
         std::cout << intp << "\n";
      }

      if( starttime. seconds_passed( ) >= maxtime )
      {
         throw std::runtime_error( "time out" );
      }

      false_instance_set current_false = findfalse( rs, intp, maxfalse ); 

      if( current_false. size( ) == 0 )
      {
         // This means that all rules are happy with all of their instances.

         return true;
      }

   use_current_false: 
      size_t has0 = 0; 
      size_t has1 = 0;
      size_t has234 = 0;
 
      for( const auto& ff : current_false )
      {
         if( ff. nrchoices == 0 )
            ++ has0;

         if( ff. nrchoices == 1 )
            ++ has1; 

         if( ff. nrchoices > 1 )
            ++ has234;  
      }

      if( has1 && !has0 )
      {

         false_instance_set dontcare{0};

         current_false. erase( []( const false_instance& f )
                                   { return f. nrchoices != 1; }, 
                                   dontcare );
           
         try_restart( current_false, dontcare );

         while( current_false. size( ))
         {
            false_instance& ff = current_false. front( );

            // Earlier deterministic extensions may change the number 
            // of choices. In particular, deterministic existential extension
            // becomes non-deterministic, as soon as an element
            // is introduced. 

            if( ff. nrchoices != 1 )
               goto findfalse;

            if( outputlevel >= 3 )
            {
               std::cout << "deterministic extension with:\n";
               std::cout << ff << "\n";
            }

            if( ff.r. isdisjunctive( ))
            {
               const disjunctive& disj = ff.r.getdisjunctive( );
                 
                  // There must be exactly one extension:

               size_t j = 0;
               while( intp. conclusionconflicts( disj, ff.inst, j ))
                  ++ j;

               intp. extend( disj, ff. inst, j, ff. assumptions );
               intp. addderivation( std::move(ff) );
            }
            else
            {
               const existential& exists = ff.r.getexistential( );

               // There must be exactly one extension. We need to find it.

               size_t j = 0;
               while( intp. conclusionconflicts( exists, ff.inst, j ))
                  ++ j;
                  
               intp. extend( exists, ff. inst, j, ff. assumptions );
               intp. addderivation( std::move(ff) );
            }

            current_false. pop_front( );
            recheck( current_false );
               // Rechecking is necessary, because, once the first
               // false formula has been extended, the others are 
               // not necessarily false anymore. In addition, 
               // deterministic rules may become non-deterministic. 
         }
         goto findfalse;
      }

      if( has0 )
      {

         {
            false_instance_set dontcare{0}; 
            current_false. erase( []( const false_instance& f )
                                    { return f. nrchoices != 0; }, dontcare );

            try_restart( current_false, dontcare );
         }

         if( outputlevel >= 3 )
         { 
            std::cout << "conflict formulas:\n";
            std::cout << current_false << "\n";
         }

         ++ intp. closure_counter;

         // In every formula of current_false, we move
         // all right hand side atoms to the left hand side,
         // in order to give them the proper format of 
         // conflict formulas. 

         for( false_instance& ff : current_false )
         {
            if( ff. r. isdisjunctive( ))
            {
               if( ff.r. getdisjunctive( ). rhs_size( ) != 0 )
               {
                  matched< disjunctive > m = 
                        { ff.r. getdisjunctive( ), ff. inst };

                  m = disj_resolve( m, {}, {} );

                  ff. r = rule( m.r, ff.r. getinference( ));
                  ff. inst = m. inst;
                  std::cout << ff << "\n\n\n";
                  throw std::runtime_error( "time to say goodbye, this is untested" ); 
               }
            }
            else
               throw std::runtime_error( "this cannot not happen" );
         }

         conflict( current_false );

      next:
         if( intp. derivations. size( ) == 0 )
         {
            false_instance_set empty{0}; 
            learn( current_false, empty ); 
            return false;  
         }
         
         // We copy the current set of conflict lemmas into 
         // the guesser:

         if( intp. nrguesses( ) == 0 )
         {
            std::cout << "derivations size " << intp. derivations. size( ) << "\n";
            std::cout << "last derivation " << intp. lastderivation( ) << "\n";
            std::cout << current_false << "\n";
            std::cout << intp << "\n";
            throw std::runtime_error( "this should never happen" );
         }

         interpretation::guesser& g = intp. lastguesser( );

         if( outputlevel >= 5 )
         {
            std::cout << "Adding conflicts into " << g << "\n";
            std::cout << "at level " << intp. nrguesses( ) - 1 << "\n";
            std::cout << current_false << "\n\n";
         }

         for( false_instance& ff : current_false )
            g. addclosinglemma( ff.r );

         // We remove the atoms that were added after the last guesser:
 
         intp. backtrackassumptions( ); 
 
         const false_instance& last = intp. lastderivation( );

         if( outputlevel >= 8 )
         {
            std::cout << "last derivation = " << last << "\n";
            std::cout << g << "\n";
            for( const auto& a : g. closing )
               std::cout << "   " << a.r << "\n";
            std::cout << "\n\n";
         }

         auto aa = last. assumptions;
         aa. insert( intp. nrguesses( ) - 1 );

         if( last. r. isexistential( ))
         {
            // We check if we have a lemma for the extension 
            // of intp with a new element: 
            //    intp + { #_{T} e, P(e) }. 
            // As soon as we have it, we derive a disjunction.

            intp. extend( last. r. getexistential( ),
                          last. inst, intp. nrelements( ), aa );

            false_instance_set ff = 
                      findconflict( g. closing, intp, maxconflict );

            auto lastelem = intp. lastelement( );

            intp. backtrackassumptions( );

            if( ff. size( ))
            {
               // We derive a disjunction and continue with this
               // disjunction in Horn, even when the disjunction
               // is not necessarily Horn. 

               current_false = 
                   resolve( { last. r. getexistential( ), last. inst },
                            userule( last.r ), lastelem, ff ); 

               false_instance_set notused{ current_false. maxsize };
               try_restart( current_false, notused ); 
               filterderived( current_false, notused );
               learn( current_false, notused );
               goto use_current_false; 
            }
         }

         size_t firstopen = g. choices. size( ); 
         size_t nropen = 0; 

         bool needmore = true;  
            // In case of disjunctive, we count all open branches.
            // In case of existential, we care only about the first. 

         for( size_t j = g.j + 1; 
                     j != g. choices. size( ) && needmore; ++ j )
         {
            if( last. r. isdisjunctive( ))
            {
               intp. extend( last. r. getdisjunctive( ),
                             last. inst, g. choices[j], aa ); 
            }
            else
            {
               intp. extend( last.r.getexistential( ), 
                             last. inst, g.choices[j], aa ); 
            }

            false_instance_set 
                     clos = findconflict( g. closing, intp, maxconflict );

            intp. backtrackassumptions( );

            if( clos. size( ) == 0 )
            {
               if( firstopen == g. choices. size( ))
                  firstopen = j;

               ++ nropen; 

               if( last. r. isexistential( ))
                  needmore = false; 
            }
         } 

         // We check how many options are still open (without conflict). 
         // We also remember the first one that is open. 

         if( last. r. isdisjunctive( ))
         { 
            if( nropen + nropen <= g. choices. size( ))
            {
               current_false = 
                  resolve( { last.r.getdisjunctive( ), last. inst },
                           userule( last.r ), g. closing );

               false_instance_set notused{ current_false. maxsize }; 
               try_restart( current_false, notused );
               filterderived( current_false, notused );
               learn( current_false, notused ); 
               goto use_current_false;  
            } 
         }

         g.j = firstopen;  
         if( last. r. isdisjunctive( ))
         {
            intp. extend( last. r. getdisjunctive( ),
                          last. inst, g. choices[ g.j ], aa );
         }
         else
         {
            intp. extend( last.r.getexistential( ),
                          last. inst, g.choices[ g.j ], aa );
         }

         goto findfalse; 
      }

      // We have only has234, so we unfortunately, we have to guess:

      while( current_false. size( ))
      {
         false_instance& ff = current_false. front( );

         if( ff. nrchoices <= 1 )
         {
            std::cout << ff << "\n";
            throw std::runtime_error( "guess what , rule become deterministic" );
         }

         if( outputlevel >= 3 )
         {
            std::cout << "guessing with:  " << ff << "\n";
         }

         // Probably impossible, because of recheck:
 
         if( ff.r. isdisjunctive( ) &&
             intp. conclusionistrue( ff.r.getdisjunctive( ), ff. inst ))
         {
            throw std::runtime_error( "conclusion became true" );
         }

         if( ff.r.isexistential( ) &&
             intp. conclusionistrue( ff.r.getexistential( ), ff. inst ))
         {
            throw std::runtime_error( "conclusion became true" );
         }

         if( ff.r. isdisjunctive( ))
         {

            // Non-deterministic extension by disjunctive rule.

            const disjunctive& disj = ff.r.getdisjunctive( );

            std::vector< size_t > extensions;
            for( size_t j = 0; j < disj. rhs_size( ); ++ j )
            {
               if( !intp. conclusionconflicts( disj, ff.inst, j )) 
                  extensions. push_back(j);
            }

            permutation perm =
            generator. makepermutation( extensions. size( ), &intp. atoms );
               
            extensions = perm( extensions ); 
         
            auto aa = ff. assumptions;
            aa. insert( intp. nrguesses( ));
 
            auto& g = intp. addguesser( std::move( extensions ));
            intp. extend( disj, ff. inst, g.choices[0], aa );
            intp. addderivation( std::move(ff) ); 
         }
         else
         {
            const existential& exists = ff.r.getexistential( );

            std::vector< size_t > extensions;
            for( size_t j = 0; j < intp. nrelements( ); ++ j )
            {
               if( !intp. conclusionconflicts( exists, ff.inst, j ))
                  extensions. push_back(j);
            }
 
            permutation perm =
            generator. makepermutation( extensions. size( ), &intp. atoms );

            extensions = perm( extensions );

            // The second last choice will extend the interpretation.
            // We add this second last choice.

            extensions. push_back( intp. nrelements( ));

            std::swap( extensions [ extensions. size( ) - 1 ],
                       extensions [ extensions. size( ) - 2 ] );
            
            auto aa = ff. assumptions;
            aa. insert( intp. nrguesses( ));
  
            auto& g = intp. addguesser( std::move( extensions ));
            intp. extend( exists, ff. inst, g.choices[0], aa );
            intp. addderivation( std::move(ff) ); 
         }

         current_false. pop_front( );
         recheck( current_false ); 
            // Rechecking is necessary, because, once the first
            // false formula has been extended, the others are not 
            // necessarily false anymore. 
      }

      goto findfalse; 

   }

}


