
// Written by Hans de Nivelle, January-July 2015.

#ifndef GEOMETRIC_SEARCHSTATE_INCLUDED
#define GEOMETRIC_SEARCHSTATE_INCLUDED   1

#include "interpretation.h"
#include "timepoints.h"
#include "rulesystem.h"
#include "atom_hasheq.h"
#include "false_instance_set.h" 
#include "deduction.h"


namespace geometric
{

   struct searchstate
   {

      permutationgenerator generator;
         // Permutation generator determines in which order 
         // branches are explored during backtracking.

      rulesystem rs; 

      age starttime;  

      interpretation intp;

      searchstate( const permutationgenerator& gen,
                   const rulesystem& rs, 
                   const std::vector< groundatom > & startatoms )
         : generator{ gen },
           rs{ rs },
           intp{ startatoms } 
      {
      } 


      searchstate( const permutationgenerator& gen,
                   rulesystem&& rs,
                   const std::vector< groundatom > & startatoms )
         : generator{ gen },
           rs{ std::move( rs ) },
           intp{ startatoms }
      {
      } 


      struct timeout 
      {
         unsigned int t;

         explicit timeout( unsigned int t ) : t{t} { } 
      };


      static constexpr size_t maxconflict = 3;
         // This is the maximum number of conflict lemmas
         // that we collect. 

      static constexpr size_t maxfalse = 3;
         // The number of false rules that we collect in one sweep
         // when we are doing forward reasoning.

      static void justderived( matched< disjunctive > & disj );
         // Sort the lhs and of the rule, and remove redundant atoms.


      static constexpr unsigned int outputlevel = 1;
         // Amount of output that we want to see.
         // 0 is very little output. 10 is very much output.

      void best_disjunctive_resolvents( 
                   const matched< disjunctive > & disj,
                   const inference& inf, 
                   const std::vector< size_t > & extensions, 
                   std::vector< false_instance > & chosen,
                   const std::vector< false_instance_set > & choices,
                   false_instance_set& best ) const;

      // The function tries all extensions of disj. It checks
      // for which extensions there are conflict lemmas. It collects
      // a maximum of maxclosing per conflict lemma.
      // After that, it backtracks through all possible combinations of 
      // closing lemmas to collect the best lemmas that can be 
      // obtained by resolution using disj and the closing lemmas. 
      // At most one conclusion of disj may have no closing lemma.
      // In that case, a Horn clause is generated.

      false_instance_set resolve( const matched< existential > & exists,
                                  const inference& inf, 
                                  element skolem_const, 
                                  false_instance_set& confl );
         // Resolve exists with each of the lemmas in confl, 
         // which should be possible, and return the results.

      false_instance_set 
      resolve( const matched< disjunctive > & disj,
               const inference& inf, 
               const std::vector< indexedrule > & conflictset );

      void conflict( false_instance_set& confl );
         // confl should contain only closing lemmas. 
         // As long as there is an extension, and this extension was 
         // obtained deterministically, we pop the last extension. 
         // If one of the lemmas in confl relies on it, we resolve. 

      void filterderived( false_instance_set& derived,
                          false_instance_set& notused );
         // Filter a set of derived formulas.

      void learn( false_instance_set& applied, 
                  false_instance_set& notapplied );
         // When a set of formulas is learnt, some of them are 
         // applied immediately, and the others are not applied. 
         // One probably should always learn the applied ones. 

      void try_restart( false_instance_set& derived, 
                        false_instance_set& notused );
         // Try to make a restart based on the false instances in derived. 
         // False_instances that are not used, are moved to notused. 
         // One can still decide to learn them if one wants.  

      void recheck( false_instance_set& inst ) const;
         // Check if the false instances in inst are still false,
         // and delete them if they are not false anymore. 
         // Recompute their number of choices. 

      bool search( unsigned int maxtime );
         // Try to extend into a model, maxtime is the maximal time 
         // in seconds. After maxtime, a time out is thrown. 
         // Return true if we find a model. If we don't find a model,
         // and no time out occured, we have found a proof. 
   };

   std::ostream& operator << ( std::ostream& out, const searchstate& ss );

}

#endif

