
// Written by Hans de Nivelle, June 2014.

#include "distribution.h"
#include <stdexcept>
#include <cmath>
#include "../tostring.h"
#include <iomanip>


statistics::distribution::distribution(  
     double s0, double s1, unsigned int nr, bool exp )
{  
   if( s0 >= s1 )
   {
      throw std::runtime_error( 
         std::string( "cannot construct distribution, because " ) +
         tostring( s0 ) + " is not less than " + tostring( s1 ));
   }
              
   if( nr == 0 )
   {
      throw std::runtime_error( 
         "cannot construct a distribution with zero steps" ); 
   }

   if( exp )
   {
      if( s0 <= 0.0 || s1 <= 0.0 )
      {
         throw std::runtime_error( 
            std::string( "cannot construct exponential distribution with " ) +
            std::string( "negative/zero start point " ) + 
            tostring( s0 )); 
      }

      double fact = pow( s1 / s0, 1.0 / nr );

      distr. push_back( 0.0 ); // To catch [ 0, s0 ).
                               // The user cannot set the lower bound. 
 
      for( unsigned int i = 0; i <= nr; ++ i )
      {
         distr. push_back( s0 );
         s0 *= fact;
      }   
   }
   else
   {
      double step = ( s1 - s0 ) / nr; 
      for( unsigned int i = 0; i <= nr; ++ i )
      {
         distr. push_back( s0 );
         s0 += step; 
      }
   }
}


void statistics::distribution::clear( ) 
{
   for( auto& e : distr )
   {
      e. sum = 0.0;
      e. nrelements = 0;
   }
}


void statistics::distribution::insert( double x )
{
   unsigned int i0 = 0;
   unsigned int i1 = distr. size( );

   // We continue search as long as there at least two elements
   // in the interval [ i0 .. i1 ).

   while( i0 + 1 < i1 )
   {
      unsigned int i = ( i0 + i1 ) / 2;
      if( distr[i]. lowerbound <= x )
         i0 = i;
      else
         i1 = i;
   }

   // We now have 0 or 1 element in interval [ i0, i1 ). 

   if( i0 == i1 )
      throw std::runtime_error( "distribution was empty, could not insert" ); 

   if( distr[i0]. lowerbound > x )
      throw std::range_error( std::string( "cannot insert " ) +
              tostring(x) + " because it is too low" );

   distr [i0]. sum += x;
   distr [i0]. nrelements += 1;
}
       

unsigned int statistics::distribution::nrelements( ) const
{
   unsigned int nr = 0;
   for( const auto& e : distr )
      nr += e. nrelements;
   return nr;
}


double statistics::distribution::sum( ) const
{
   double s = 0.0;
   for( const auto& e : distr )
      s += e. sum;
   return s;
}


std::ostream& statistics::operator << ( std::ostream& stream,
		                        const distribution& d )
{
   double partial_sum = 0.0;
   unsigned int partial_nrelements = 0; 

   double total_sum = d. sum( );
   unsigned int total_nrelements = d. nrelements( );

   stream << "interval                       ";
   stream << "nrelements (cumulative %)";
   stream << "      average (cumulative sum %)\n";
   stream << "---------------------------------------------";
   stream << "----------------------------------------\n";

   for( unsigned int i = 0; i < d. distr. size( ); ++ i )
   {
      if( d. distr[i]. nrelements )
      {
         stream << "[ " << std::setw(10) << d. distr[i]. lowerbound;

         if( i + 1 < d. distr. size( ))
            stream << ", " << std::setw(10) << 
                           d. distr[i+1]. lowerbound << " ) : ";
         else
            stream << ", ....      ) : ";  

         double this_sum = d. distr[i]. sum; 
         unsigned int this_nrelements = d. distr[i]. nrelements;  

         stream << "  ";
         stream << std::setw(10) << this_nrelements; 

         partial_nrelements += this_nrelements;
         stream << "  (";
         stream << std::setw(10) << 
                   static_cast<double>( partial_nrelements ) / 
                   static_cast<double>( total_nrelements );
         stream << ")";

         stream << "     ";
         stream. width(10); 
         stream << ( this_sum / this_nrelements ); 

         stream << "  (";

         partial_sum += this_sum; 
         stream. width(10); 
         stream << partial_sum / total_sum; 
         stream << ")\n"; 
      }
   }

   if( total_nrelements )
   {
      stream << "total                          ";
      stream. width(10); 
      stream << total_nrelements; 
      stream << "                   ";
      stream. width(10); 
      stream << total_sum / total_nrelements;

      stream << "\n\n";
   }
   return stream;
}



